@@ -17,6 +17,8 @@ static const unsigned char fentry_nop_patt2[] = { 0x0f, 0x1f, 0x44, 0x00, 0x00 }
 static const unsigned char patchable_gcc_nop[] = { 0x90, 0x90, 0x90, 0x90, 0x90 };
 static const unsigned char patchable_clang_nop[] = { 0x0f, 0x1f, 0x44, 0x00, 0x08 };
 
+static const unsigned char endbr64[] = { 0xf3, 0x0f, 0x1e, 0xfa };
+
 int mcount_setup_trampoline(struct mcount_dynamic_info *mdi)
 {
 	unsigned char trampoline[] = { 0x3e, 0xff, 0x25, 0x01, 0x00, 0x00, 0x00, 0xcc };
@@ -143,21 +145,21 @@ static void read_patchable_loc(struct mcount_dynamic_info *mdi, struct uftrace_e
 			       struct uftrace_elf_iter *iter, unsigned long offset)
 {
 	typeof(iter->shdr) *shdr = &iter->shdr;
+	unsigned i;
+	unsigned long *patchable_loc;
+	unsigned long sh_addr;
 
 	mdi->nr_patch_target = shdr->sh_size / sizeof(long);
 	mdi->patch_target = xmalloc(shdr->sh_size);
+	patchable_loc = mdi->patch_target;
 
-	elf_get_secdata(elf, iter);
-	elf_read_secdata(elf, iter, 0, mdi->patch_target, shdr->sh_size);
+	sh_addr = shdr->sh_addr;
+	if (elf->ehdr.e_type == ET_DYN)
+		sh_addr += offset;
 
-	/* symbol has relative address, fix it to match each other */
-	if (elf->ehdr.e_type == ET_EXEC) {
-		unsigned long *patchable_loc = mdi->patch_target;
-		unsigned i;
-
-		for (i = 0; i < mdi->nr_patch_target; i++) {
-			patchable_loc[i] -= offset;
-		}
+	for (i = 0; i < mdi->nr_patch_target; i++) {
+		unsigned long *entry = (unsigned long *)sh_addr + i;
+		patchable_loc[i] = *entry - offset;
 	}
 }
 
@@ -255,6 +257,10 @@ static int patch_fentry_code(struct mcount_dynamic_info *mdi, struct uftrace_sym
 	unsigned char *insn = (void *)sym->addr + mdi->map->start;
 	unsigned int target_addr;
 
+	/* skip 'endbr64' instruction, which is inserted by (implicit) -fcf-protection option. */
+	if (!memcmp(insn, endbr64, sizeof(endbr64)))
+		insn += sizeof(endbr64);
+
 	/* support patchable function entry and __fentry__ at the beginning */
 	if (memcmp(insn, patchable_gcc_nop, sizeof(patchable_gcc_nop)) &&
 	    memcmp(insn, patchable_clang_nop, sizeof(patchable_clang_nop)) &&
@@ -634,7 +640,6 @@ static void revert_normal_func(struct mcount_dynamic_info *mdi, struct uftrace_s
 			       struct mcount_disasm_engine *disasm)
 {
 	void *addr = (void *)(uintptr_t)sym->addr + mdi->map->start;
-	uint8_t endbr64[] = { 0xf3, 0x0f, 0x1e, 0xfa };
 	struct mcount_orig_insn *moi;
 
 	if (!memcmp(addr, endbr64, sizeof(endbr64)))
