@@ -536,7 +536,8 @@ static void dump_raw_header(struct uftrace_dump_ops *ops, struct uftrace_data *h
 	struct uftrace_raw_dump *raw = container_of(ops, typeof(*raw), ops);
 	const char *feat_str[] = { "PLTHOOK",	 "TASK_SESSION", "KERNEL",     "ARGUMENT",
 				   "RETVAL",	 "SYM_REL_ADDR", "MAX_STACK",  "EVENT",
-				   "PERF_EVENT", "AUTO_ARGS",	 "DEBUG_INFO", "ESTIMATE_RETURN" };
+				   "PERF_EVENT", "AUTO_ARGS",	 "DEBUG_INFO", "ESTIMATE_RETURN",
+				   "SYM_SIZE" };
 	const char *info_str[] = { "EXE_NAME",	   "EXE_BUILD_ID", "EXIT_STATUS", "CMDLINE",
 				   "CPUINFO",	   "MEMINFO",	   "OSINFO",	  "TASKINFO",
 				   "USAGEINFO",	   "LOADINFO",	   "ARG_SPEC",	  "RECORD_DATE",
@@ -1344,7 +1345,7 @@ static void dump_mermaid_footer(struct uftrace_dump_ops *ops, struct uftrace_dat
 	pr_out("</h2>\n");
 
 	interactive_option_html =
-#include "utils/mermaid.html.cstr" /* This file is a converted of mermaid.html to one string literal in build-time */
+#include "utils/mermaid.html.cstr" /* This file is converted from mermaid.html at build-time */
 		;
 
 	pr_out("%s", interactive_option_html);
@@ -1364,7 +1365,7 @@ static void dump_mermaid_footer(struct uftrace_dump_ops *ops, struct uftrace_dat
 	pr_out("mermaid.initialize(config);\n");
 
 	interactive_option_js =
-#include "utils/mermaid.js.cstr" /* This file is a converted of mermaid.html to one string literal in build-time */
+#include "utils/mermaid.js.cstr" /* This file is converted from mermaid.js at build-time */
 		;
 
 	pr_out("%s", interactive_option_js);
