@@ -25,6 +25,7 @@
 #include "utils/kernel.h"
 #include "utils/list.h"
 #include "utils/perf.h"
+#include "utils/shmem.h"
 #include "utils/symbol.h"
 #include "utils/utils.h"
 
@@ -298,8 +299,8 @@ static void setup_child_environ(struct uftrace_opts *opts, int argc, char *argv[
 		setenv("UFTRACE_DEBUG_DOMAIN", build_debug_domain_string(), 1);
 	}
 
-	if (opts->disabled)
-		setenv("UFTRACE_DISABLED", "1", 1);
+	if (opts->trace == TRACE_STATE_OFF)
+		setenv("UFTRACE_TRACE_OFF", "1", 1);
 
 	if (log_color == COLOR_ON) {
 		snprintf(buf, sizeof(buf), "%d", log_color);
@@ -413,6 +414,9 @@ static uint64_t calc_feat_mask(struct uftrace_opts *opts)
 	if (opts->estimate_return)
 		features |= ESTIMATE_RETURN;
 
+	/* symbol file saves size */
+	features |= SYM_SIZE;
+
 	xasprintf(&buf, "%s/*.dbg", opts->dirname);
 	if (glob(buf, GLOB_NOSORT, NULL, &g) != GLOB_NOMATCH)
 		features |= DEBUG_INFO;
@@ -823,7 +827,7 @@ static void record_mmap_file(const char *dirname, char *sess_id, int bufsize)
 	struct mcount_shmem_buffer *shmem_buf;
 
 	/* write (append) it to disk */
-	fd = shm_open(sess_id, O_RDWR, 0600);
+	fd = uftrace_shmem_open(sess_id, O_RDWR, 0600);
 	if (fd < 0) {
 		pr_dbg("open shmem buffer failed: %s: %m\n", sess_id);
 		return;
@@ -932,12 +936,12 @@ static void unlink_shmem_list(void)
 		sscanf(sl->id, "/uftrace-%[^-]-%*d-%*d", shmem_session);
 		pr_dbg2("unlink for session: %s\n", shmem_session);
 
-		num = scandir("/dev/shm/", &shmem_bufs, filter_shmem, alphasort);
+		num = scandir(uftrace_shmem_root(), &shmem_bufs, filter_shmem, alphasort);
 		for (i = 0; i < num; i++) {
 			sid[0] = '/';
 			memcpy(&sid[1], shmem_bufs[i]->d_name, MSG_ID_SIZE);
 			pr_dbg3("unlink %s\n", sid);
-			shm_unlink(sid);
+			uftrace_shmem_unlink(sid);
 			free(shmem_bufs[i]);
 		}
 
@@ -1547,35 +1551,32 @@ static bool is_regular_executable(const char *pathname)
 	return false;
 }
 
-static void find_in_path(char **exename)
+static void find_in_path(char *exename, char *buf, size_t len)
 {
 	/* try to find the binary in PATH */
 	struct strv strv = STRV_INIT;
 	char *env = getenv("PATH");
-	char *pathname = NULL;
 	char *path;
 	bool found = false;
 	int i;
 
-	if (!env || *exename[0] == '/')
-		pr_err_ns(UFTRACE_MSG, *exename);
+	if (!env || exename[0] == '/')
+		pr_err_ns(UFTRACE_MSG, exename);
 
 	/* search opts->exename in PATH one by one */
 	strv_split(&strv, env, ":");
 
 	strv_for_each(&strv, path, i) {
-		xasprintf(&pathname, "%s/%s", path, *exename);
-		if (is_regular_executable(pathname)) {
+		snprintf(buf, len, "%s/%s", path, exename);
+		if (is_regular_executable(buf)) {
 			found = true;
 			break;
 		}
-		free(pathname);
 	}
 
 	if (!found)
-		pr_err_ns(UFTRACE_MSG, *exename);
+		pr_err_ns(UFTRACE_MSG, exename);
 
-	*exename = pathname;
 	strv_free(&strv);
 }
 
@@ -1585,14 +1586,17 @@ static void check_binary(struct uftrace_opts *opts)
 	int chk;
 	size_t i;
 	char elf_ident[EI_NIDENT];
+	static char altname[PATH_MAX]; // for opts->exename to be persistent
 	uint16_t e_type;
 	uint16_t e_machine;
 	uint16_t supported_machines[] = { EM_X86_64, EM_ARM, EM_AARCH64, EM_386 };
 
 again:
 	/* if it cannot be found in PATH, then fails inside */
-	if (!is_regular_executable(opts->exename))
-		find_in_path(&opts->exename);
+	if (!is_regular_executable(opts->exename)) {
+		find_in_path(opts->exename, altname, sizeof(altname));
+		opts->exename = altname;
+	}
 
 	pr_dbg("checking binary %s\n", opts->exename);
 
@@ -1604,15 +1608,25 @@ again:
 		pr_err("Cannot read '%s'", opts->exename);
 
 	if (memcmp(elf_ident, ELFMAG, SELFMAG)) {
-		char *script = check_script_file(opts->exename);
+		char *script = altname;
 		char *p;
 
-		if (script == NULL)
+		if (!check_script_file(opts->exename, altname, sizeof(altname)))
 			pr_err_ns(UFTRACE_ELF_MSG, opts->exename);
 
+#if defined(HAVE_LIBPYTHON2) || defined(HAVE_LIBPYTHON3)
+		if (strstr(script, "python")) {
+			opts->force = true;
+			/* TODO: disable sched event until it can merge subsequent events */
+			opts->no_sched = true;
+		}
+#endif
+
 		if (!opts->force && !opts->patch)
 			pr_err_ns(SCRIPT_MSG, opts->exename);
 
+		script = str_ltrim(script);
+
 		/* ignore options */
 		p = strchr(script, ' ');
 		if (p)
@@ -2124,9 +2138,11 @@ int do_main_loop(int ready, struct uftrace_opts *opts, int pid)
 int do_child_exec(int ready, struct uftrace_opts *opts, int argc, char *argv[])
 {
 	uint64_t dummy;
-	char *shebang;
-	char fullpath[PATH_MAX];
+	char *shebang = NULL;
+	char dirpath[PATH_MAX];
+	char exepath[PATH_MAX];
 	struct strv new_args = STRV_INIT;
+	bool is_python = false;
 
 	if (opts->no_randomize_addr) {
 		/* disable ASLR (Address Space Layout Randomization) */
@@ -2138,18 +2154,41 @@ int do_child_exec(int ready, struct uftrace_opts *opts, int argc, char *argv[])
 	 * The current working directory can be changed by calling chdir.
 	 * So dirname has to be converted to an absolute path to avoid unexpected problems.
 	 */
-	if (realpath(opts->dirname, fullpath) != NULL)
-		opts->dirname = fullpath;
+	if (realpath(opts->dirname, dirpath) != NULL)
+		opts->dirname = dirpath;
+
+	if (access(argv[0], F_OK) == 0) {
+		/* prefer current directory over PATH */
+		if (check_script_file(argv[0], exepath, sizeof(exepath)))
+			shebang = exepath;
+	}
+	else {
+		struct strv path_names = STRV_INIT;
+		char *path, *dir;
+		int i, ret;
+
+		strv_split(&path_names, getenv("PATH"), ":");
+		strv_for_each(&path_names, dir, i) {
+			xasprintf(&path, "%s/%s", dir, argv[0]);
+			ret = access(path, F_OK);
+			if (ret == 0 && check_script_file(path, exepath, sizeof(exepath)))
+				shebang = exepath;
+			free(path);
+			if (ret == 0)
+				break;
+		}
+		strv_free(&path_names);
+	}
 
-	shebang = check_script_file(argv[0]);
 	if (shebang) {
 		char *s, *p;
 		int i;
 
-		s = shebang;
-
-		while (isspace(*s))
-			s++;
+#if defined(HAVE_LIBPYTHON2) || defined(HAVE_LIBPYTHON3)
+		if (strstr(shebang, "python"))
+			is_python = true;
+#endif
+		s = str_ltrim(shebang);
 
 		p = strchr(s, ' ');
 		if (p != NULL)
@@ -2159,13 +2198,22 @@ int do_child_exec(int ready, struct uftrace_opts *opts, int argc, char *argv[])
 		if (p != NULL)
 			strv_append(&new_args, p);
 
+		if (is_python) {
+			strv_append(&new_args, "-m");
+			strv_append(&new_args, "uftrace");
+			if (!opts->libcall)
+				setenv("UFTRACE_PY_LIBCALL", "NONE", 1);
+			if (opts->nest_libcall)
+				setenv("UFTRACE_PY_LIBCALL", "NESTED", 1);
+			/* disable library calls for 'python' interpreter */
+			opts->libcall = false;
+		}
+
 		for (i = 0; i < argc; i++)
 			strv_append(&new_args, argv[i]);
 
 		argc = new_args.nr;
 		argv = new_args.p;
-
-		free(shebang);
 	}
 
 	setup_child_environ(opts, argc, argv);
@@ -2174,6 +2222,26 @@ int do_child_exec(int ready, struct uftrace_opts *opts, int argc, char *argv[])
 	if (read(ready, &dummy, sizeof(dummy)) != (ssize_t)sizeof(dummy))
 		pr_err("waiting for parent failed");
 
+	if (is_python) {
+		char *python_path = NULL;
+
+		if (getenv("PYTHONPATH"))
+			python_path = strdup(getenv("PYTHONPATH"));
+
+#ifdef INSTALL_LIB_PATH
+		python_path = strjoin(python_path, INSTALL_LIB_PATH, ":");
+#endif
+		python_path = strjoin(python_path, "python", ":"); /* FIXME */
+		setenv("PYTHONPATH", python_path, 1);
+		free(python_path);
+
+		/*
+		 * prevent from creating .pyc files inside __pycache__.
+		 * it makes some script execution failed.
+		 */
+		setenv("PYTHONDONTWRITEBYTECODE", "1", 1);
+	}
+
 	/*
 	 * The traced binary is already resolved into absolute pathname.
 	 * So plain 'execv' is enough and no need to use 'execvp'.
