@@ -1,32 +1,31 @@
-#include <stdio.h>
-#include <stdlib.h>
-#include <signal.h>
 #include <inttypes.h>
+#include <signal.h>
+#include <stdio.h>
 #include <stdio_ext.h>
-#include <assert.h>
+#include <stdlib.h>
 
 #include "uftrace.h"
-#include "utils/utils.h"
-#include "utils/symbol.h"
+#include "utils/event.h"
+#include "utils/field.h"
 #include "utils/filter.h"
 #include "utils/fstack.h"
-#include "utils/list.h"
 #include "utils/kernel.h"
-#include "utils/field.h"
+#include "utils/list.h"
+#include "utils/symbol.h"
+#include "utils/utils.h"
 
 #include "libtraceevent/event-parse.h"
 
-
 static int column_index;
 static int prev_tid = -1;
 
 static LIST_HEAD(output_fields);
 
-#define NO_TIME  (void *)1  /* to suppress duration */
+#define NO_TIME (void *)1 /* to suppress duration */
 
 static void print_duration(struct field_data *fd)
 {
-	struct fstack *fstack = fd->fstack;
+	struct uftrace_fstack *fstack = fd->fstack;
 	void *arg = fd->arg;
 	uint64_t d = 0;
 
@@ -45,25 +44,25 @@ static void print_tid(struct field_data *fd)
 
 static void print_addr(struct field_data *fd)
 {
-	struct fstack *fstack = fd->fstack;
+	struct uftrace_fstack *fstack = fd->fstack;
 
 	/* uftrace records (truncated) 48-bit addresses */
 	int width = sizeof(long) == 4 ? 8 : 12;
 
-	if (fstack == NULL)  /* LOST */
+	if (fstack == NULL) /* LOST */
 		pr_out("%*s", width, "");
 	else
-		pr_out("%*"PRIx64, width, effective_addr(fstack->addr));
+		pr_out("%*" PRIx64, width, effective_addr(fstack->addr));
 }
 
 static void print_timestamp(struct field_data *fd)
 {
 	struct uftrace_task_reader *task = fd->task;
 
-	uint64_t  sec = task->timestamp / NSEC_PER_SEC;
+	uint64_t sec = task->timestamp / NSEC_PER_SEC;
 	uint64_t nsec = task->timestamp % NSEC_PER_SEC;
 
-	pr_out("%8"PRIu64".%09"PRIu64, sec, nsec);
+	pr_out("%8" PRIu64 ".%09" PRIu64, sec, nsec);
 }
 
 static void print_timedelta(struct field_data *fd)
@@ -95,7 +94,7 @@ static void print_task(struct field_data *fd)
 static void print_module(struct field_data *fd)
 {
 	struct uftrace_task_reader *task = fd->task;
-	struct fstack *fstack = fd->fstack;
+	struct uftrace_fstack *fstack = fd->fstack;
 	uint64_t timestamp = task->timestamp;
 	struct uftrace_session *s;
 	struct uftrace_mmap *map;
@@ -109,107 +108,102 @@ static void print_module(struct field_data *fd)
 
 	s = find_task_session(&task->h->sessions, task->t, timestamp);
 	if (s) {
-		map = find_map(&s->symtabs, fstack->addr);
+		map = find_map(&s->sym_info, fstack->addr);
 		if (map == MAP_KERNEL)
 			modname = "[kernel]";
 		else if (map)
 			modname = basename(map->libname);
+		else if (is_sched_event(fstack->addr))
+			modname = "[event]";
 	}
 
 	pr_out("%*.*s", 16, 16, modname);
 }
 
 static struct display_field field_duration = {
-	.id      = REPLAY_F_DURATION,
-	.name    = "duration",
-	.header  = " DURATION ",
-	.length  = 10,
-	.print   = print_duration,
-	.list    = LIST_HEAD_INIT(field_duration.list),
+	.id = REPLAY_F_DURATION,
+	.name = "duration",
+	.header = " DURATION ",
+	.length = 10,
+	.print = print_duration,
+	.list = LIST_HEAD_INIT(field_duration.list),
 };
 
 static struct display_field field_tid = {
-	.id      = REPLAY_F_TID,
-	.name    = "tid",
-	.header  = "   TID  ",
-	.length  = 8,
-	.print   = print_tid,
-	.list    = LIST_HEAD_INIT(field_tid.list),
+	.id = REPLAY_F_TID,
+	.name = "tid",
+	.header = "   TID  ",
+	.length = 8,
+	.print = print_tid,
+	.list = LIST_HEAD_INIT(field_tid.list),
 };
 
 static struct display_field field_addr = {
-	.id      = REPLAY_F_ADDR,
-	.name    = "addr",
+	.id = REPLAY_F_ADDR,
+	.name = "addr",
 #if __SIZEOF_LONG__ == 4
-	.header  = " ADDRESS",
-	.length  = 8,
+	.header = " ADDRESS",
+	.length = 8,
 #else
-	.header  = "   ADDRESS  ",
-	.length  = 12,
+	.header = "   ADDRESS  ",
+	.length = 12,
 #endif
-	.print   = print_addr,
-	.list    = LIST_HEAD_INIT(field_addr.list),
+	.print = print_addr,
+	.list = LIST_HEAD_INIT(field_addr.list),
 };
 
 static struct display_field field_time = {
-	.id      = REPLAY_F_TIMESTAMP,
-	.name    = "time",
-	.header  = "     TIMESTAMP    ",
-	.length  = 18,
-	.print   = print_timestamp,
-	.list    = LIST_HEAD_INIT(field_time.list),
+	.id = REPLAY_F_TIMESTAMP,
+	.name = "time",
+	.header = "     TIMESTAMP    ",
+	.length = 18,
+	.print = print_timestamp,
+	.list = LIST_HEAD_INIT(field_time.list),
 };
 
 static struct display_field field_delta = {
-	.id      = REPLAY_F_TIMEDELTA,
-	.name    = "delta",
-	.header  = " TIMEDELTA",
-	.length  = 10,
-	.print   = print_timedelta,
-	.list    = LIST_HEAD_INIT(field_delta.list),
+	.id = REPLAY_F_TIMEDELTA,
+	.name = "delta",
+	.header = " TIMEDELTA",
+	.length = 10,
+	.print = print_timedelta,
+	.list = LIST_HEAD_INIT(field_delta.list),
 };
 
 static struct display_field field_elapsed = {
-	.id      = REPLAY_F_ELAPSED,
-	.name    = "elapsed",
-	.header  = "  ELAPSED ",
-	.length  = 10,
-	.print   = print_elapsed,
-	.list    = LIST_HEAD_INIT(field_elapsed.list),
+	.id = REPLAY_F_ELAPSED,
+	.name = "elapsed",
+	.header = "  ELAPSED ",
+	.length = 10,
+	.print = print_elapsed,
+	.list = LIST_HEAD_INIT(field_elapsed.list),
 };
 
 static struct display_field field_task = {
-	.id      = REPLAY_F_TASK,
-	.name    = "task",
-	.header  = "      TASK NAME",
-	.length  = 15,
-	.print   = print_task,
-	.list    = LIST_HEAD_INIT(field_task.list),
+	.id = REPLAY_F_TASK,
+	.name = "task",
+	.header = "      TASK NAME",
+	.length = 15,
+	.print = print_task,
+	.list = LIST_HEAD_INIT(field_task.list),
 };
 
 static struct display_field field_module = {
-	.id      = REPLAY_F_MODULE,
-	.name    = "module",
-	.header  = "     MODULE NAME",
-	.length  = 16,
-	.print   = print_module,
-	.list    = LIST_HEAD_INIT(field_module.list),
+	.id = REPLAY_F_MODULE,
+	.name = "module",
+	.header = "     MODULE NAME",
+	.length = 16,
+	.print = print_module,
+	.list = LIST_HEAD_INIT(field_module.list),
 };
 
 /* index of this table should be matched to display_field_id */
 static struct display_field *field_table[] = {
-	&field_duration,
-	&field_tid,
-	&field_addr,
-	&field_time,
-	&field_delta,
-	&field_elapsed,
-	&field_task,
-	&field_module,
+	&field_duration, &field_tid,	 &field_addr, &field_time,
+	&field_delta,	 &field_elapsed, &field_task, &field_module,
 };
 
-static void print_field(struct uftrace_task_reader *task,
-			struct fstack *fstack, void *arg)
+static void print_field(struct uftrace_task_reader *task, struct uftrace_fstack *fstack, void *arg)
 {
 	struct field_data fd = {
 		.task = task,
@@ -221,7 +215,8 @@ static void print_field(struct uftrace_task_reader *task,
 		pr_out(" | ");
 }
 
-static void setup_default_field(struct list_head *fields, struct opts *opts)
+static void setup_default_field(struct list_head *fields, struct uftrace_opts *opts,
+				struct display_field *p_field_table[])
 {
 	if (opts->range.start > 0 || opts->range.stop > 0) {
 		if (opts->range.start_elapsed || opts->range.stop_elapsed)
@@ -233,7 +228,7 @@ static void setup_default_field(struct list_head *fields, struct opts *opts)
 	add_field(fields, field_table[REPLAY_F_TID]);
 }
 
-static int task_column_depth(struct uftrace_task_reader *task, struct opts *opts)
+static int task_column_depth(struct uftrace_task_reader *task, struct uftrace_opts *opts)
 {
 	if (!opts->column_view)
 		return 0;
@@ -251,9 +246,9 @@ static void print_backtrace(struct uftrace_task_reader *task)
 
 	for (i = 0; i < task->stack_count - 1; i++) {
 		struct display_field *field;
-		struct sym *sym;
+		struct uftrace_symbol *sym;
 		char *name;
-		struct fstack *fstack = fstack_get(task, i);
+		struct uftrace_fstack *fstack = fstack_get(task, i);
 		struct field_data fd = {
 			.task = task,
 			.fstack = fstack,
@@ -262,8 +257,7 @@ static void print_backtrace(struct uftrace_task_reader *task)
 		if (fstack == NULL)
 			continue;
 
-		sym = task_find_sym_addr(sessions, task,
-					 fstack->total_time, fstack->addr);
+		sym = task_find_sym_addr(sessions, task, fstack->total_time, fstack->addr);
 
 		pr_out(" ");
 		list_for_each_entry(field, &output_fields, list) {
@@ -282,12 +276,11 @@ static void print_backtrace(struct uftrace_task_reader *task)
 	}
 }
 
-static void print_event(struct uftrace_task_reader *task,
-			struct uftrace_record *urec,
-			int color)
+static void print_event(struct uftrace_task_reader *task, struct uftrace_record *urec, int color)
 {
 	unsigned evt_id = urec->addr;
-	char *evt_name = get_event_name(task->h, evt_id);
+	char *evt_name = event_get_name(task->h, evt_id);
+	char *evt_data = event_get_data_str(evt_id, task->args.data, true);
 
 	if (evt_id == EVENT_ID_EXTERN_DATA) {
 		pr_color(color, "%s: %s", evt_name, (char *)task->args.data);
@@ -296,102 +289,26 @@ static void print_event(struct uftrace_task_reader *task,
 		/* TODO: some events might have arguments */
 		pr_color(color, "%s", evt_name);
 	}
-	else if (evt_id >= EVENT_ID_PERF) {
+	else {
 		pr_color(color, "%s", evt_name);
 
-		if (evt_id == EVENT_ID_PERF_COMM)
-			pr_color(color, " (name=%s)", (char *)task->args.data);
+		if (evt_data)
+			pr_color(color, " (%s)", evt_data);
 	}
-	else if (evt_id >= EVENT_ID_BUILTIN) {
-		union {
-			struct uftrace_proc_statm *statm;
-			struct uftrace_page_fault *page_fault;
-			struct uftrace_pmu_cycle  *cycle;
-			struct uftrace_pmu_cache  *cache;
-			struct uftrace_pmu_branch *branch;
-			int                       *cpu;
-		} u;
 
-		switch (evt_id) {
-		case EVENT_ID_READ_PROC_STATM:
-			u.statm = task->args.data;
-			pr_color(color, "%s (size=%"PRIu64"KB, rss=%"PRIu64"KB, shared=%"PRIu64"KB)",
-				 evt_name, u.statm->vmsize, u.statm->vmrss, u.statm->shared);
-			break;
-		case EVENT_ID_READ_PAGE_FAULT:
-			u.page_fault = task->args.data;
-			pr_color(color, "%s (major=%"PRIu64", minor=%"PRIu64")",
-				 evt_name, u.page_fault->major, u.page_fault->minor);
-			break;
-		case EVENT_ID_READ_PMU_CYCLE:
-			u.cycle = task->args.data;
-			pr_color(color, "%s (cycle=%"PRIu64", instructions=%"PRIu64")",
-				 evt_name, u.cycle->cycles, u.cycle->instrs);
-			break;
-		case EVENT_ID_READ_PMU_CACHE:
-			u.cache = task->args.data;
-			pr_color(color, "%s (refers=%"PRIu64", misses=%"PRIu64")",
-				 evt_name, u.cache->refers, u.cache->misses);
-			break;
-		case EVENT_ID_READ_PMU_BRANCH:
-			u.branch = task->args.data;
-			pr_color(color, "%s (branch=%"PRIu64", misses=%"PRIu64")",
-				 evt_name, u.branch->branch, u.branch->misses);
-			break;
-		case EVENT_ID_DIFF_PROC_STATM:
-			u.statm = task->args.data;
-			pr_color(color, "%s (size=%+"PRId64"KB, rss=%+"PRId64"KB, shared=%+"PRId64"KB)",
-				 evt_name, u.statm->vmsize, u.statm->vmrss, u.statm->shared);
-			break;
-		case EVENT_ID_DIFF_PAGE_FAULT:
-			u.page_fault = task->args.data;
-			pr_color(color, "%s (major=%+"PRId64", minor=%+"PRId64")",
-				 evt_name, u.page_fault->major, u.page_fault->minor);
-			break;
-		case EVENT_ID_DIFF_PMU_CYCLE:
-			u.cycle = task->args.data;
-			pr_color(color, "%s (cycle=%+"PRId64", instructions=%+"PRId64", IPC=%.2f)",
-				 evt_name, u.cycle->cycles, u.cycle->instrs,
-				 (float)u.cycle->instrs / u.cycle->cycles);
-			break;
-		case EVENT_ID_DIFF_PMU_CACHE:
-			u.cache = task->args.data;
-			pr_color(color, "%s (refers=%+"PRId64", misses=%+"PRId64", hit=%"PRId64"%%)",
-				 evt_name, u.cache->refers, u.cache->misses,
-				 (u.cache->refers - u.cache->misses) * 100 / u.cache->refers);
-			break;
-		case EVENT_ID_DIFF_PMU_BRANCH:
-			u.branch = task->args.data;
-			pr_color(color, "%s (branch=%+"PRId64", misses=%+"PRId64", predict=%"PRId64"%%)",
-				 evt_name, u.branch->branch, u.branch->misses,
-				 (u.branch->branch - u.branch->misses) * 100 / u.branch->branch);
-			break;
-		case EVENT_ID_WATCH_CPU:
-			u.cpu = task->args.data;
-			pr_color(color, "%s (cpu=%d)", evt_name, *u.cpu);
-			break;
-		default:
-			pr_color(color, "%s", evt_name);
-		}
-	}
-	else {
-		/* kernel events */
-		pr_color(color, "%s (%.*s)", evt_name,
-			 task->args.len, task->args.data);
-	}
 	free(evt_name);
+	free(evt_data);
 }
 
-static int print_flat_rstack(struct uftrace_data *handle,
-			     struct uftrace_task_reader *task,
-			     struct opts *opts)
+static int print_flat_rstack(struct uftrace_data *handle, struct uftrace_task_reader *task,
+			     struct uftrace_opts *opts)
 {
 	static int count;
 	struct uftrace_record *rstack = task->rstack;
 	struct uftrace_session_link *sessions = &task->h->sessions;
-	struct sym *sym = NULL;
+	struct uftrace_symbol *sym = NULL;
 	char *name;
-	struct fstack *fstack;
+	struct uftrace_fstack *fstack;
 
 	sym = task_find_sym(sessions, task, rstack);
 	name = symbol_getname(sym, rstack->addr);
@@ -406,26 +323,23 @@ static int print_flat_rstack(struct uftrace_data *handle,
 
 	switch (rstack->type) {
 	case UFTRACE_ENTRY:
-		pr_out("[%d] ==> %d/%d: ip (%s), time (%"PRIu64")\n",
-		       count++, task->tid, rstack->depth,
-		       name, rstack->time);
+		pr_out("[%d] ==> %d/%d: ip (%s), time (%" PRIu64 ")\n", count++, task->tid,
+		       rstack->depth, name, rstack->time);
 		break;
 
 	case UFTRACE_EXIT:
-		pr_out("[%d] <== %d/%d: ip (%s), time (%"PRIu64":%"PRIu64")\n",
-		       count++, task->tid, rstack->depth,
-		       name, rstack->time, fstack->total_time);
+		pr_out("[%d] <== %d/%d: ip (%s), time (%" PRIu64 ":%" PRIu64 ")\n", count++,
+		       task->tid, rstack->depth, name, rstack->time, fstack->total_time);
 		break;
 
 	case UFTRACE_LOST:
-		pr_out("[%d] XXX %d: lost %d records\n",
-		       count++, task->tid, (int)rstack->addr);
+		pr_out("[%d] XXX %d: lost %d records\n", count++, task->tid, (int)rstack->addr);
 		break;
 
 	case UFTRACE_EVENT:
 		pr_out("[%d] !!! %d: ", count++, task->tid);
 		print_event(task, rstack, task->event_color);
-		pr_out(" time (%"PRIu64")\n", rstack->time);
+		pr_out(" time (%" PRIu64 ")\n", rstack->time);
 		break;
 	}
 out:
@@ -444,43 +358,55 @@ static void print_task_newline(int current_tid)
 	prev_tid = current_tid;
 }
 
-#define print_char(c)                                                   \
-({ args[n] = c; n++; len--; })
+static void print_char(char **args, size_t *len, const char c)
+{
+	**args = c;
+	*args += 1;
+	*len -= 1;
+}
 
-#define print_args(fmt, ...)						\
-({ int _x = snprintf(args + n, len, fmt, ##__VA_ARGS__); n += _x; len -= _x; })
+static void print_args(char **args, size_t *len, const char *fmt, ...)
+{
+	int x;
+	va_list ap;
 
-#define print_json_escaped_char(c)					\
-	do {								\
-		if (c == '\n')						\
-			print_args("\\\\n");				\
-		else if (c == '\t')					\
-			print_args("\\\\t");				\
-		else if (c == '\\')					\
-			print_args("\\\\");				\
-		else if (c == '"')					\
-			print_args("\\\"");				\
-		else if (isprint(c))					\
-			print_char(c);                                  \
-		else							\
-			print_args("\\\\x%02hhx", c);			\
-	} while (0)
+	va_start(ap, fmt);
+	x = vsnprintf(*args, *len, fmt, ap);
+	va_end(ap);
+	*args += x;
+	*len -= x;
+}
 
-#define print_escaped_char(c)						\
-	do {								\
-		if (c == '\0')						\
-			print_args("\\0");				\
-		else if (c == '\b')					\
-			print_args("\\b");				\
-		else if (c == '\n')                                     \
-			print_args("\\n");				\
-		else							\
-			print_char(c);                                  \
-	} while (0)
+void print_json_escaped_char(char **args, size_t *len, const char c)
+{
+	if (c == '\n')
+		print_args(args, len, "\\\\n");
+	else if (c == '\t')
+		print_args(args, len, "\\\\t");
+	else if (c == '\\')
+		print_args(args, len, "\\\\");
+	else if (c == '"')
+		print_args(args, len, "\\\"");
+	else if (isprint(c))
+		print_char(args, len, c);
+	else
+		print_args(args, len, "\\\\x%02hhx", c);
+}
 
-void get_argspec_string(struct uftrace_task_reader *task,
-			char *args, size_t len,
-			enum argspec_string_bits str_mode)
+static void print_escaped_char(char **args, size_t *len, const char c)
+{
+	if (c == '\0')
+		print_args(args, len, "\\0");
+	else if (c == '\b')
+		print_args(args, len, "\\b");
+	else if (c == '\n')
+		print_args(args, len, "\\n");
+	else
+		print_char(args, len, c);
+}
+
+void get_argspec_string(struct uftrace_task_reader *task, char *args, size_t len,
+			enum uftrace_argspec_string_bits str_mode)
 {
 	int i = 0, n = 0;
 	char *str = NULL;
@@ -499,12 +425,12 @@ void get_argspec_string(struct uftrace_task_reader *task,
 		unsigned char v[16];
 	} val;
 
-	bool needs_paren      = !!(str_mode & NEEDS_PAREN);
+	bool needs_paren = !!(str_mode & NEEDS_PAREN);
 	bool needs_semi_colon = !!(str_mode & NEEDS_SEMI_COLON);
-	bool has_more         = !!(str_mode & HAS_MORE);
-	bool is_retval        = !!(str_mode & IS_RETVAL);
+	bool has_more = !!(str_mode & HAS_MORE);
+	bool is_retval = !!(str_mode & IS_RETVAL);
 	bool needs_assignment = !!(str_mode & NEEDS_ASSIGNMENT);
-	bool needs_json       = !!(str_mode & NEEDS_JSON);
+	bool needs_json = !!(str_mode & NEEDS_JSON);
 
 	if (!has_more) {
 		if (needs_paren)
@@ -517,12 +443,12 @@ void get_argspec_string(struct uftrace_task_reader *task,
 		return;
 	}
 
-	assert(arg_list && !list_empty(arg_list));
+	ASSERT(arg_list && !list_empty(arg_list));
 
 	if (needs_paren)
-		print_args("%s(%s", color_bold, color_reset);
+		print_args(&args, &len, "(");
 	else if (needs_assignment)
-		print_args("%s = %s", color_bold, color_reset);
+		print_args(&args, &len, " = ");
 
 	list_for_each_entry(spec, arg_list, list) {
 		char fmtstr[16];
@@ -536,7 +462,7 @@ void get_argspec_string(struct uftrace_task_reader *task,
 			continue;
 
 		if (i > 0)
-			print_args("%s, %s", color_bold, color_reset);
+			print_args(&args, &len, ", ");
 
 		memset(val.v, 0, sizeof(val));
 		fmt = ARG_SPEC_CHARS[spec->fmt];
@@ -555,8 +481,7 @@ void get_argspec_string(struct uftrace_task_reader *task,
 				if (sizeof(long) == sizeof(uint64_t)) {
 					uint64_t val64 = val.i;
 
-					if (val64 >  0xffff0000 &&
-					    val64 <= 0xffffffff) {
+					if (val64 > 0xffff0000 && val64 <= 0xffffffff) {
 						fmt = 'd';
 						idx = 2;
 						break;
@@ -579,8 +504,7 @@ void get_argspec_string(struct uftrace_task_reader *task,
 			break;
 		}
 
-		if (spec->fmt == ARG_FMT_STR ||
-		    spec->fmt == ARG_FMT_STD_STRING) {
+		if (spec->fmt == ARG_FMT_STR || spec->fmt == ARG_FMT_STD_STRING) {
 			unsigned short slen;
 
 			memcpy(&slen, data, 2);
@@ -590,21 +514,21 @@ void get_argspec_string(struct uftrace_task_reader *task,
 			str[slen] = '\0';
 
 			if (slen == 4 && !memcmp(str, &null_str, sizeof(null_str)))
-				print_args("NULL");
+				print_args(&args, &len, "NULL");
 			else if (needs_json) {
 				char *p = str;
-				print_args("\\\"");
+
+				print_args(&args, &len, "\\\"");
 				while (*p) {
 					char c = *p++;
-					print_json_escaped_char(c);
+					print_json_escaped_char(&args, &len, c);
 				}
-				print_args("\\\"");
+				print_args(&args, &len, "\\\"");
 			}
 			else {
-				print_args("%s", color_string);
-				print_args("\"");
-
 				char *p = str;
+
+				print_args(&args, &len, "%s\"", color_string);
 				while (*p) {
 					char c = *p++;
 					if (c & 0x80) {
@@ -616,23 +540,22 @@ void get_argspec_string(struct uftrace_task_reader *task,
 				* then it will be UTF-8 string
 				*/
 				if (*p) {
-					print_args("%.*s", slen, str);
+					print_args(&args, &len, "%.*s", slen, str);
 				}
 				else {
 					p = str;
 					while (*p) {
 						char c = *p++;
-						print_escaped_char(c);
+						print_escaped_char(&args, &len, c);
 					}
 				}
 
-				print_args("\"");
-				print_args("%s", color_reset);
+				print_args(&args, &len, "\"%s", color_reset);
 			}
 
 			/* std::string can be represented as "TEXT"s from C++14 */
 			if (spec->fmt == ARG_FMT_STD_STRING)
-				print_args("s");
+				print_args(&args, &len, "s");
 
 			free(str);
 			size = slen + 2;
@@ -642,16 +565,16 @@ void get_argspec_string(struct uftrace_task_reader *task,
 
 			memcpy(&c, data, 1);
 			if (needs_json) {
-				print_args("'");
-				print_json_escaped_char(c);
-				print_args("'");
+				print_args(&args, &len, "'");
+				print_json_escaped_char(&args, &len, c);
+				print_args(&args, &len, "'");
 			}
 			else {
-				print_args("%s", color_string);
-				print_args("'");
-				print_escaped_char(c);
-				print_args("'");
-				print_args("%s", color_reset);
+				print_args(&args, &len, "%s", color_string);
+				print_args(&args, &len, "'");
+				print_escaped_char(&args, &len, c);
+				print_args(&args, &len, "'");
+				print_args(&args, &len, "%s", color_reset);
 			}
 			size = 1;
 		}
@@ -666,58 +589,58 @@ void get_argspec_string(struct uftrace_task_reader *task,
 
 			switch (spec->size) {
 			case 4:
-				print_args(fmtstr, val.f);
+				print_args(&args, &len, fmtstr, val.f);
 				break;
 			case 8:
-				print_args(fmtstr, val.d);
+				print_args(&args, &len, fmtstr, val.d);
 				break;
 			case 10:
-				print_args(fmtstr, val.D);
+				print_args(&args, &len, fmtstr, val.D);
 				break;
 			default:
-				pr_dbg("invalid floating-point type size %d\n",
-				       spec->size);
+				pr_dbg("invalid floating-point type size %d\n", spec->size);
 				break;
 			}
 		}
 		else if (spec->fmt == ARG_FMT_PTR) {
 			struct uftrace_session_link *sessions = &task->h->sessions;
-			struct sym *sym;
+			struct uftrace_symbol *sym;
 
 			memcpy(val.v, data, spec->size);
-			sym = task_find_sym_addr(sessions, task,
-						 task->rstack->time,
+			sym = task_find_sym_addr(sessions, task, task->rstack->time,
 						 (uint64_t)val.i);
 
 			if (sym) {
-				print_args("%s", color_symbol);
-				print_args("&%s", sym->name);
-				print_args("%s", color_reset);
+				print_args(&args, &len, "%s", color_symbol);
+				if (format_mode == FORMAT_HTML)
+					print_args(&args, &len, "&amp;%s", sym->name);
+				else
+					print_args(&args, &len, "&%s", sym->name);
+				print_args(&args, &len, "%s", color_reset);
 			}
 			else if (val.p)
-				print_args("%p", val.p);
+				print_args(&args, &len, "%p", val.p);
 			else
-				print_args("0");
+				print_args(&args, &len, "0");
 		}
 		else if (spec->fmt == ARG_FMT_ENUM) {
 			struct uftrace_session_link *sessions = &task->h->sessions;
 			struct uftrace_session *s;
 			struct uftrace_mmap *map;
-			struct debug_info *dinfo;
+			struct uftrace_dbg_info *dinfo;
 			char *estr;
 
 			memcpy(val.v, data, spec->size);
-			s = find_task_session(sessions, task->t,
-					      task->rstack->time);
+			s = find_task_session(sessions, task->t, task->rstack->time);
 
-			map = find_map(&s->symtabs, task->rstack->addr);
+			map = find_map(&s->sym_info, task->rstack->addr);
 			if (map == NULL || map->mod == NULL) {
-				print_args("<ENUM?> %lx", val.i);
+				print_args(&args, &len, "<ENUM?> %x", (int)val.i);
 				goto next;
 			}
 
 			dinfo = &map->mod->dinfo;
-			estr = get_enum_string(&dinfo->enums, spec->enum_str, val.i);
+			estr = get_enum_string(&dinfo->enums, spec->type_name, (int)val.i);
 			if (strstr(estr, "|") && strcmp("|", color_enum_or)) {
 				struct strv enum_vals = STRV_INIT;
 
@@ -727,26 +650,42 @@ void get_argspec_string(struct uftrace_task_reader *task,
 				strv_free(&enum_vals);
 			}
 
-			print_args("%s", color_enum);
+			print_args(&args, &len, "%s", color_enum);
 			if (strlen(estr) >= len)
-				print_args("<ENUM>");
+				print_args(&args, &len, "<ENUM>");
 			else
-				print_args("%s", estr);
-			print_args("%s", color_reset);
+				print_args(&args, &len, "%s", estr);
+			print_args(&args, &len, "%s", color_reset);
 			free(estr);
 		}
+		else if (spec->fmt == ARG_FMT_STRUCT) {
+			if (spec->type_name) {
+				/*
+				 * gcc puts "<lambda" to anonymous lambda
+				 * but let's ignore to make it same as clang.
+				 */
+				if (strcmp(spec->type_name, "<lambda")) {
+					print_args(&args, &len, "%s%s%s", color_struct,
+						   spec->type_name, color_reset);
+				}
+			}
+			if (spec->size)
+				print_args(&args, &len, "{...}");
+			else
+				print_args(&args, &len, "{}");
+		}
 		else {
 			if (spec->fmt != ARG_FMT_AUTO)
 				memcpy(val.v, data, spec->size);
 
-			assert(idx < ARRAY_SIZE(len_mod));
+			ASSERT(idx < ARRAY_SIZE(len_mod));
 			lm = len_mod[idx];
 
 			snprintf(fmtstr, sizeof(fmtstr), "%%#%s%c", lm, fmt);
 			if (spec->size == 8)
-				print_args(fmtstr, val.ll);
+				print_args(&args, &len, fmtstr, val.ll);
 			else
-				print_args(fmtstr, val.i);
+				print_args(&args, &len, fmtstr, val.i);
 		}
 
 next:
@@ -762,26 +701,28 @@ next:
 	}
 
 	if (needs_paren) {
-		print_args("%s)%s", color_bold, color_reset);
-	} else {
+		print_args(&args, &len, ")");
+	}
+	else {
 		if (needs_semi_colon)
 			args[n++] = ';';
 		args[n] = '\0';
 	}
 }
 
-static int print_graph_rstack(struct uftrace_data *handle,
-			      struct uftrace_task_reader *task,
-			      struct opts *opts)
+static int print_graph_rstack(struct uftrace_data *handle, struct uftrace_task_reader *task,
+			      struct uftrace_opts *opts)
 {
 	struct uftrace_record *rstack;
 	struct uftrace_session_link *sessions = &handle->sessions;
-	struct sym *sym = NULL;
-	enum argspec_string_bits str_mode = 0;
+	struct uftrace_symbol *sym = NULL;
+	enum uftrace_argspec_string_bits str_mode = 0;
 	char *symname = NULL;
 	char args[1024];
 	char *libname = "";
 	struct uftrace_mmap *map = NULL;
+	struct uftrace_dbg_loc *loc = NULL;
+	char *str_loc = NULL;
 
 	if (task == NULL)
 		return 0;
@@ -810,15 +751,21 @@ static int print_graph_rstack(struct uftrace_data *handle,
 
 		s = find_task_session(sessions, task->t, rstack->time);
 		if (s != NULL) {
-			map = find_symbol_map(&s->symtabs, symname);
+			map = find_symbol_map(&s->sym_info, symname);
 			if (map != NULL)
 				libname = basename(map->libname);
 		}
 	}
 
+	if (opts->srcline) {
+		loc = task_find_loc_addr(sessions, task, rstack->time, rstack->addr);
+		if (opts->comment && loc)
+			xasprintf(&str_loc, "%s:%d", loc->file->name, loc->line);
+	}
+
 	if (rstack->type == UFTRACE_ENTRY) {
 		struct uftrace_task_reader *next = NULL;
-		struct fstack *fstack;
+		struct uftrace_fstack *fstack;
 		int rstack_depth = rstack->depth;
 		int depth;
 		struct uftrace_trigger tr = {
@@ -847,7 +794,7 @@ static int print_graph_rstack(struct uftrace_data *handle,
 
 		depth += task_column_depth(task, opts);
 
-		if (rstack->more)
+		if (rstack->more && opts->show_args)
 			str_mode |= HAS_MORE;
 		get_argspec_string(task, args, sizeof(args), str_mode);
 
@@ -856,8 +803,7 @@ static int print_graph_rstack(struct uftrace_data *handle,
 		if (!opts->no_merge)
 			next = fstack_skip(handle, task, rstack_depth, opts);
 
-		if (task == next &&
-		    next->rstack->depth == rstack_depth &&
+		if (task == next && next->rstack->depth == rstack_depth &&
 		    next->rstack->type == UFTRACE_EXIT) {
 			char retval[1024];
 
@@ -865,7 +811,7 @@ static int print_graph_rstack(struct uftrace_data *handle,
 			fstack_consume(handle, next);
 
 			str_mode = IS_RETVAL | NEEDS_SEMI_COLON;
-			if (next->rstack->more) {
+			if (next->rstack->more && opts->show_args) {
 				str_mode |= HAS_MORE;
 				str_mode |= NEEDS_ASSIGNMENT;
 			}
@@ -877,13 +823,15 @@ static int print_graph_rstack(struct uftrace_data *handle,
 				pr_color(tr.color, "%s", symname);
 				if (*libname)
 					pr_color(tr.color, "@%s", libname);
-				pr_out("%s%s\n", args, retval);
+				pr_out("%s%s", args, retval);
 			}
 			else {
-				pr_out("%s%s%s%s%s\n", symname,
-				       *libname ? "@" : "",
-				       libname, args, retval);
+				pr_out("%s%s%s%s%s", symname, *libname ? "@" : "", libname, args,
+				       retval);
 			}
+			if (str_loc)
+				pr_gray(" /* %s */", str_loc);
+			pr_out("\n");
 
 			/* fstack_update() is not needed here */
 
@@ -897,31 +845,32 @@ static int print_graph_rstack(struct uftrace_data *handle,
 				pr_color(tr.color, "%s", symname);
 				if (*libname)
 					pr_color(tr.color, "@%s", libname);
-				pr_out("%s {\n", args);
+				pr_out("%s {", args);
 			}
 			else {
-				pr_out("%s%s%s%s {\n", symname,
-				       *libname ? "@" : "", libname, args);
+				pr_out("%s%s%s%s {", symname, *libname ? "@" : "", libname, args);
 			}
+			if (str_loc)
+				pr_gray(" /* %s */", str_loc);
+			pr_out("\n");
 
 			fstack_update(UFTRACE_ENTRY, task, fstack);
 		}
 	}
 	else if (rstack->type == UFTRACE_EXIT) {
-		struct fstack *fstack;
+		struct uftrace_fstack *fstack;
 
 		/* function exit */
 		fstack = fstack_get(task, task->stack_count);
 
-		if (fstack_enabled && fstack != NULL &&
-		    !(fstack->flags & FSTACK_FL_NORECORD)) {
+		if (fstack_enabled && fstack != NULL && !(fstack->flags & FSTACK_FL_NORECORD)) {
 			int depth = fstack_update(UFTRACE_EXIT, task, fstack);
 			char *retval = args;
 
 			depth += task_column_depth(task, opts);
 
 			str_mode = IS_RETVAL;
-			if (rstack->more) {
+			if (rstack->more && opts->show_args) {
 				str_mode |= HAS_MORE;
 				str_mode |= NEEDS_ASSIGNMENT;
 				str_mode |= NEEDS_SEMI_COLON;
@@ -934,11 +883,10 @@ static int print_graph_rstack(struct uftrace_data *handle,
 
 			print_field(task, fstack, NULL);
 			pr_out("%*s}%s", depth * 2, "", retval);
-			if (opts->comment)
-				pr_gray(" /* %s%s%s */\n", symname,
-					*libname ? "@" : "", libname);
-			else
-				pr_gray("\n");
+			if (opts->comment) {
+				pr_gray(" /* %s%s%s */", symname, *libname ? "@" : "", libname);
+			}
+			pr_out("\n");
 		}
 
 		fstack_exit(task);
@@ -960,16 +908,15 @@ lost:
 		print_field(task, NULL, NO_TIME);
 
 		if (losts > 0)
-			pr_red("%*s/* LOST %d records!! */\n",
-			       depth * 2, "", losts);
+			pr_red("%*s/* LOST %d records!! */\n", depth * 2, "", losts);
 		else /* kernel sometimes have unknown count */
-			pr_red("%*s/* LOST some records!! */\n",
-			       depth * 2, "");
+			pr_red("%*s/* LOST some records!! */\n", depth * 2, "");
+		free(str_loc);
 		return 0;
 	}
 	else if (rstack->type == UFTRACE_EVENT) {
 		int depth;
-		struct fstack *fstack;
+		struct uftrace_fstack *fstack;
 		struct uftrace_task_reader *next = NULL;
 		struct uftrace_record rec = *rstack;
 		uint64_t evt_id = rstack->addr;
@@ -989,15 +936,21 @@ lost:
 		 * try to merge a subsequent sched-in event:
 		 * it might overwrite rstack - use (saved) rec for printing.
 		 */
-		if (evt_id == EVENT_ID_PERF_SCHED_OUT && !opts->no_merge)
+		if ((evt_id == EVENT_ID_PERF_SCHED_OUT ||
+		     evt_id == EVENT_ID_PERF_SCHED_OUT_PREEMPT) &&
+		    !opts->no_merge)
 			next = fstack_skip(handle, task, 0, opts);
 
-		if (task == next &&
-		    next->rstack->addr == EVENT_ID_PERF_SCHED_IN) {
+		if (task == next && next->rstack->addr == EVENT_ID_PERF_SCHED_IN) {
 			/* consume the matching sched-in record */
 			fstack_consume(handle, next);
 
-			rec.addr = sched_sym.addr;
+			if (evt_id == EVENT_ID_PERF_SCHED_OUT)
+				rec.addr = EVENT_ID_PERF_SCHED_BOTH;
+			else if (evt_id == EVENT_ID_PERF_SCHED_OUT_PREEMPT)
+				rec.addr = EVENT_ID_PERF_SCHED_BOTH_PREEMPT;
+			else
+				pr_warn("unexpected event id\n");
 			evt_id = EVENT_ID_PERF_SCHED_IN;
 		}
 
@@ -1008,10 +961,8 @@ lost:
 		/* for sched-in to show schedule duration */
 		fstack = fstack_get(task, task->stack_count);
 
-		if (fstack_enabled && fstack != NULL &&
-		    !(fstack->flags & FSTACK_FL_NORECORD)) {
-			if (evt_id == EVENT_ID_PERF_SCHED_IN &&
-			    fstack->total_time)
+		if (fstack_enabled && fstack != NULL && !(fstack->flags & FSTACK_FL_NORECORD)) {
+			if (evt_id == EVENT_ID_PERF_SCHED_IN && fstack->total_time)
 				print_field(task, fstack, NULL);
 			else
 				print_field(task, NULL, NO_TIME);
@@ -1020,10 +971,10 @@ lost:
 			print_event(task, &rec, task->event_color);
 			pr_color(task->event_color, " */\n");
 		}
-
 	}
 out:
 	symbol_putname(sym, symname);
+	free(str_loc);
 	return 0;
 }
 
@@ -1031,14 +982,13 @@ static void print_warning(struct uftrace_task_reader *task)
 {
 	if (print_empty_field(&output_fields, 1))
 		pr_out(" | ");
-	pr_red(" %*s/* inverted time: broken data? */\n",
-	       (task->display_depth + 1) * 2, "");
+	pr_red(" %*s/* inverted time: broken data? */\n", (task->display_depth + 1) * 2, "");
 }
 
-static bool skip_sys_exit(struct opts *opts, struct uftrace_task_reader *task)
+static bool skip_sys_exit(struct uftrace_opts *opts, struct uftrace_task_reader *task)
 {
-	struct sym *sym;
-	struct fstack *fstack;
+	struct uftrace_symbol *sym;
+	struct uftrace_fstack *fstack;
 
 	fstack = fstack_get(task, 0);
 	if (fstack == NULL)
@@ -1048,7 +998,7 @@ static bool skip_sys_exit(struct opts *opts, struct uftrace_task_reader *task)
 	if (!has_kernel_data(task->h->kernel) || task->user_stack_count != 0)
 		return false;
 
-	sym = find_symtabs(&task->h->sessions.first->symtabs, fstack->addr);
+	sym = find_symtabs(&task->h->sessions.first->sym_info, fstack->addr);
 	if (sym == NULL)
 		return false;
 
@@ -1061,8 +1011,7 @@ static bool skip_sys_exit(struct opts *opts, struct uftrace_task_reader *task)
 	return false;
 }
 
-static void print_remaining_stack(struct opts *opts,
-				  struct uftrace_data *handle)
+static void print_remaining_stack(struct uftrace_opts *opts, struct uftrace_data *handle)
 {
 	int i, k;
 	int total = 0;
@@ -1075,8 +1024,18 @@ static void print_remaining_stack(struct opts *opts,
 		if (skip_sys_exit(opts, task))
 			continue;
 
+		if (task->stack_count == 1) {
+			struct uftrace_fstack *fstack = fstack_get(task, 0);
+
+			/* ignore if it only has a schedule event */
+			if (fstack && (fstack->addr == EVENT_ID_PERF_SCHED_OUT ||
+				       fstack->addr == EVENT_ID_PERF_SCHED_OUT_PREEMPT))
+				continue;
+		}
+
+		/* sometimes it has many 0 entries in the fstack. ignore them */
 		for (k = 0; k < task->stack_count; k++) {
-			struct fstack *fstack;
+			struct uftrace_fstack *fstack;
 
 			fstack = fstack_get(task, k);
 			if (fstack != NULL && fstack->addr != 0)
@@ -1095,12 +1054,21 @@ static void print_remaining_stack(struct opts *opts,
 
 	for (i = 0; i < handle->nr_tasks; i++) {
 		struct uftrace_task_reader *task = &handle->tasks[i];
-		struct fstack *fstack;
+		struct uftrace_fstack *fstack;
 		int zero_count = 0;
 
 		if (task->stack_count == 0)
 			continue;
 
+		if (task->stack_count == 1) {
+			fstack = fstack_get(task, 0);
+
+			/* skip if it only has a schedule event */
+			if (fstack && (fstack->addr == EVENT_ID_PERF_SCHED_OUT ||
+				       fstack->addr == EVENT_ID_PERF_SCHED_OUT_PREEMPT))
+				continue;
+		}
+
 		for (k = 0; k < task->stack_count; k++) {
 			fstack = fstack_get(task, k);
 			if (fstack != NULL && fstack->addr != 0)
@@ -1119,7 +1087,7 @@ static void print_remaining_stack(struct opts *opts,
 		while (task->stack_count-- > 0) {
 			uint64_t time;
 			uint64_t ip;
-			struct sym *sym;
+			struct uftrace_symbol *sym;
 			char *symname;
 
 			fstack = fstack_get(task, task->stack_count);
@@ -1142,7 +1110,7 @@ static void print_remaining_stack(struct opts *opts,
 	}
 }
 
-int command_replay(int argc, char *argv[], struct opts *opts)
+int command_replay(int argc, char *argv[], struct uftrace_opts *opts)
 {
 	int ret;
 	uint64_t prev_time = 0;
@@ -1159,11 +1127,19 @@ int command_replay(int argc, char *argv[], struct opts *opts)
 	}
 
 	fstack_setup_filters(opts, &handle);
-	setup_field(&output_fields, opts, &setup_default_field,
-		    field_table, ARRAY_SIZE(field_table));
+	setup_field(&output_fields, opts, &setup_default_field, field_table,
+		    ARRAY_SIZE(field_table));
+
+	if (format_mode == FORMAT_HTML)
+		pr_out(HTML_HEADER);
 
 	if (!opts->flat && peek_rstack(&handle, &task) == 0)
-		print_header(&output_fields, "#", "FUNCTION", 1);
+		print_header(&output_fields, "#", "FUNCTION", 1, false);
+	if (!list_empty(&output_fields)) {
+		if (opts->srcline)
+			pr_gray(" [SOURCE]");
+		pr_out("\n");
+	}
 
 	while (read_rstack(&handle, &task) == 0 && !uftrace_done) {
 		struct uftrace_record *rstack = task->rstack;
@@ -1194,6 +1170,9 @@ int command_replay(int argc, char *argv[], struct opts *opts)
 
 	print_remaining_stack(opts, &handle);
 
+	if (format_mode == FORMAT_HTML)
+		pr_out(HTML_FOOTER);
+
 	close_data_file(opts, &handle);
 
 	return ret;
