@@ -76,24 +76,22 @@ int page_size_in_kb;
 /* call depth to filter */
 static int __maybe_unused mcount_depth = MCOUNT_DEFAULT_DEPTH;
 
+/* setting for all filter actions */
+static struct uftrace_filter_setting mcount_filter_setting = {
+	.ptype = PATT_REGEX,
+	.auto_args = false,
+	.allow_kernel = false,
+};
+
 /* boolean flag to turn on/off recording */
 static bool __maybe_unused mcount_enabled = true;
 
-/* function filtering mode - inclusive or exclusive */
-static enum filter_mode __maybe_unused mcount_filter_mode = FILTER_MODE_NONE;
-
-/* location filtering mode - inclusive or exclusive */
-static enum filter_mode __maybe_unused mcount_loc_mode = FILTER_MODE_NONE;
-
-/* tree of trigger actions */
-static struct rb_root __maybe_unused mcount_triggers = RB_ROOT;
+/* triggers definition and counters */
+static struct uftrace_triggers_info __maybe_unused *mcount_triggers;
 
 /* bitmask of active watch points */
 static unsigned long __maybe_unused mcount_watchpoints;
 
-/* whether caller filter is activated */
-static bool __maybe_unused mcount_has_caller;
-
 /* address of function will be called when a function returns */
 unsigned long mcount_return_fn;
 
@@ -104,7 +102,12 @@ bool mcount_estimate_return;
 static pthread_t agent;
 
 /* state flag for the agent */
-static bool agent_run = false;
+static volatile bool agent_run = false;
+
+#define MCOUNT_AGENT_CAPABILITIES                                                                  \
+	(UFTRACE_AGENT_OPT_TRACE | UFTRACE_AGENT_OPT_DEPTH | UFTRACE_AGENT_OPT_THRESHOLD |         \
+	 UFTRACE_AGENT_OPT_PATTERN | UFTRACE_AGENT_OPT_FILTER | UFTRACE_AGENT_OPT_CALLER |         \
+	 UFTRACE_AGENT_OPT_TRIGGER)
 
 __weak void dynamic_return(void)
 {
@@ -112,7 +115,7 @@ __weak void dynamic_return(void)
 
 #ifdef DISABLE_MCOUNT_FILTER
 
-static void mcount_filter_init(enum uftrace_pattern_type ptype, bool force)
+static void mcount_filter_init(struct uftrace_filter_setting *filter_setting, bool force)
 {
 	if (getenv("UFTRACE_SRCLINE") == NULL)
 		return;
@@ -120,7 +123,7 @@ static void mcount_filter_init(enum uftrace_pattern_type ptype, bool force)
 	load_module_symtabs(&mcount_sym_info);
 
 	/* use debug info if available */
-	prepare_debug_info(&mcount_sym_info, ptype, NULL, NULL, false, force);
+	prepare_debug_info(&mcount_sym_info, filter_setting->ptype, NULL, NULL, false, force);
 	save_debug_info(&mcount_sym_info, mcount_sym_info.dirname);
 }
 
@@ -364,7 +367,7 @@ static void mcount_signal_finish(void)
 	}
 }
 
-static void mcount_filter_init(enum uftrace_pattern_type ptype, bool force)
+static void mcount_filter_init(struct uftrace_filter_setting *filter_setting, bool force)
 {
 	char *filter_str = getenv("UFTRACE_FILTER");
 	char *trigger_str = getenv("UFTRACE_TRIGGER");
@@ -373,24 +376,19 @@ static void mcount_filter_init(enum uftrace_pattern_type ptype, bool force)
 	char *autoargs_str = getenv("UFTRACE_AUTO_ARGS");
 	char *caller_str = getenv("UFTRACE_CALLER");
 	char *loc_str = getenv("UFTRACE_LOCATION");
-
-	struct uftrace_filter_setting filter_setting = {
-		.ptype = ptype,
-		.auto_args = false,
-		.allow_kernel = false,
-		.lp64 = host_is_lp64(),
-		.arch = host_cpu_arch(),
-	};
 	bool needs_debug_info = false;
 
+	filter_setting->lp64 = host_is_lp64();
+	filter_setting->arch = host_cpu_arch();
+
 	load_module_symtabs(&mcount_sym_info);
 
-	mcount_signal_init(getenv("UFTRACE_SIGNAL"), &filter_setting);
+	mcount_signal_init(getenv("UFTRACE_SIGNAL"), filter_setting);
 
 	/* setup auto-args only if argument/return value is used */
 	if (argument_str || retval_str || autoargs_str ||
 	    (trigger_str && (strstr(trigger_str, "arg") || strstr(trigger_str, "retval")))) {
-		setup_auto_args(&filter_setting);
+		setup_auto_args(filter_setting);
 		needs_debug_info = true;
 	}
 
@@ -399,57 +397,54 @@ static void mcount_filter_init(enum uftrace_pattern_type ptype, bool force)
 
 	/* use debug info if available */
 	if (needs_debug_info) {
-		prepare_debug_info(&mcount_sym_info, ptype, argument_str, retval_str,
-				   !!autoargs_str, force);
+		prepare_debug_info(&mcount_sym_info, filter_setting->ptype, argument_str,
+				   retval_str, !!autoargs_str, force);
 		save_debug_info(&mcount_sym_info, mcount_sym_info.dirname);
 	}
 
-	uftrace_setup_filter(filter_str, &mcount_sym_info, &mcount_triggers, &mcount_filter_mode,
-			     &filter_setting);
-	uftrace_setup_trigger(trigger_str, &mcount_sym_info, &mcount_triggers, &mcount_filter_mode,
-			      &filter_setting);
-	uftrace_setup_argument(argument_str, &mcount_sym_info, &mcount_triggers, &filter_setting);
-	uftrace_setup_retval(retval_str, &mcount_sym_info, &mcount_triggers, &filter_setting);
+	mcount_triggers = xmalloc(sizeof(*mcount_triggers));
+	memset(mcount_triggers, 0, sizeof(*mcount_triggers));
+	mcount_triggers->root = RB_ROOT;
+	uftrace_setup_filter(filter_str, &mcount_sym_info, mcount_triggers, filter_setting);
+	uftrace_setup_trigger(trigger_str, &mcount_sym_info, mcount_triggers, filter_setting);
+	uftrace_setup_argument(argument_str, &mcount_sym_info, mcount_triggers, filter_setting);
+	uftrace_setup_retval(retval_str, &mcount_sym_info, mcount_triggers, filter_setting);
 
 	if (needs_debug_info) {
-		uftrace_setup_loc_filter(loc_str, &mcount_sym_info, &mcount_triggers,
-					 &mcount_loc_mode, &filter_setting);
+		uftrace_setup_loc_filter(loc_str, &mcount_sym_info, mcount_triggers,
+					 filter_setting);
 	}
 
 	if (caller_str) {
-		uftrace_setup_caller_filter(caller_str, &mcount_sym_info, &mcount_triggers,
-					    &filter_setting);
+		uftrace_setup_caller_filter(caller_str, &mcount_sym_info, mcount_triggers,
+					    filter_setting);
 	}
 
-	/* there might be caller triggers, count it separately */
-	if (uftrace_count_filter(&mcount_triggers, TRIGGER_FL_CALLER) != 0)
-		mcount_has_caller = true;
-
 	if (autoargs_str) {
 		char *autoarg = ".";
 		char *autoret = ".";
 
-		if (ptype == PATT_GLOB)
+		if (filter_setting->ptype == PATT_GLOB)
 			autoarg = autoret = "*";
 
-		filter_setting.auto_args = true;
+		filter_setting->auto_args = true;
 
-		uftrace_setup_argument(autoarg, &mcount_sym_info, &mcount_triggers,
-				       &filter_setting);
-		uftrace_setup_retval(autoret, &mcount_sym_info, &mcount_triggers, &filter_setting);
+		uftrace_setup_argument(autoarg, &mcount_sym_info, mcount_triggers, filter_setting);
+		uftrace_setup_retval(autoret, &mcount_sym_info, mcount_triggers, filter_setting);
 	}
 
 	if (getenv("UFTRACE_DEPTH"))
 		mcount_depth = strtol(getenv("UFTRACE_DEPTH"), NULL, 0);
 
-	if (getenv("UFTRACE_DISABLED"))
+	if (getenv("UFTRACE_TRACE_OFF"))
 		mcount_enabled = false;
 }
 
 static void mcount_filter_setup(struct mcount_thread_data *mtdp)
 {
-	mtdp->filter.depth = mcount_depth;
-	mtdp->filter.time = mcount_threshold;
+	mtdp->filter.max_depth = FILTER_NO_MAX_DEPTH;
+	mtdp->filter.depth = 0;
+	mtdp->filter.time = FILTER_NO_TIME;
 	mtdp->filter.size = mcount_min_size;
 	mtdp->enable_cached = mcount_enabled;
 	mtdp->argbuf = xmalloc(mcount_rstack_max * ARGBUF_SIZE);
@@ -465,7 +460,8 @@ static void mcount_filter_release(struct mcount_thread_data *mtdp)
 
 static void mcount_filter_finish(void)
 {
-	uftrace_cleanup_filter(&mcount_triggers);
+	uftrace_cleanup_triggers(mcount_triggers);
+	free(mcount_triggers);
 	finish_auto_args();
 
 	finish_debug_info(&mcount_sym_info);
@@ -843,10 +839,27 @@ static bool mcount_check_rstack(struct mcount_thread_data *mtdp)
 #ifndef DISABLE_MCOUNT_FILTER
 extern void *get_argbuf(struct mcount_thread_data *, struct mcount_ret_stack *);
 
+/**
+ * mcount_get_filter_mode - compute the filter mode from the filter count
+ */
+static inline enum filter_mode mcount_get_filter_mode(void)
+{
+	return mcount_triggers->filter_count > 0 ? FILTER_MODE_IN : FILTER_MODE_OUT;
+}
+
+/**
+ * mcount_get_loc_mode - compute the location filter mode from the location count
+ */
+static inline enum filter_mode mcount_get_loc_mode(void)
+{
+	return mcount_triggers->loc_count > 0 ? FILTER_MODE_IN : FILTER_MODE_OUT;
+}
+
 static void mcount_save_filter(struct mcount_thread_data *mtdp)
 {
 	/* save original depth and time to restore at exit time */
 	mtdp->filter.saved_depth = mtdp->filter.depth;
+	mtdp->filter.saved_max_depth = mtdp->filter.max_depth;
 	mtdp->filter.saved_time = mtdp->filter.time;
 	mtdp->filter.saved_size = mtdp->filter.size;
 }
@@ -855,8 +868,12 @@ static void mcount_save_filter(struct mcount_thread_data *mtdp)
 enum filter_result mcount_entry_filter_check(struct mcount_thread_data *mtdp, unsigned long child,
 					     struct uftrace_trigger *tr)
 {
-	pr_dbg3("<%d> enter %lx\n", mtdp->idx, child);
+	int max_depth = mtdp->filter.max_depth;
 
+	if (max_depth == FILTER_NO_MAX_DEPTH)
+		max_depth = mcount_depth;
+
+	pr_dbg3("<%d> enter %lx\n", mtdp->idx, child);
 	if (mcount_check_rstack(mtdp))
 		return FILTER_RSTACK;
 
@@ -866,7 +883,7 @@ enum filter_result mcount_entry_filter_check(struct mcount_thread_data *mtdp, un
 	if (mtdp->filter.out_count > 0)
 		return FILTER_OUT;
 
-	uftrace_match_filter(child, &mcount_triggers, tr);
+	uftrace_match_filter(child, &mcount_triggers->root, tr);
 
 	pr_dbg3(" tr->flags: %x, filter mode: %d, count: %d/%d, depth: %d\n", tr->flags, tr->fmode,
 		mtdp->filter.in_count, mtdp->filter.out_count, mtdp->filter.depth);
@@ -878,11 +895,11 @@ enum filter_result mcount_entry_filter_check(struct mcount_thread_data *mtdp, un
 			mtdp->filter.out_count++;
 
 		/* apply default filter depth when match */
-		mtdp->filter.depth = mcount_depth;
+		mtdp->filter.depth = 0;
 	}
 	else {
 		/* not matched by filter */
-		if (mcount_filter_mode == FILTER_MODE_IN && mtdp->filter.in_count == 0)
+		if (mcount_get_filter_mode() == FILTER_MODE_IN && mtdp->filter.in_count == 0)
 			return FILTER_OUT;
 	}
 
@@ -891,21 +908,20 @@ enum filter_result mcount_entry_filter_check(struct mcount_thread_data *mtdp, un
 			return FILTER_OUT;
 	}
 	else {
-		if (mcount_loc_mode == FILTER_MODE_IN)
+		if (mcount_get_loc_mode() == FILTER_MODE_IN)
 			return FILTER_OUT;
 	}
 
 #define FLAGS_TO_CHECK                                                                             \
 	(TRIGGER_FL_DEPTH | TRIGGER_FL_TRACE_ON | TRIGGER_FL_TRACE_OFF | TRIGGER_FL_TIME_FILTER |  \
 	 TRIGGER_FL_SIZE_FILTER)
-
 	if (tr->flags & FLAGS_TO_CHECK) {
-		if (tr->flags & TRIGGER_FL_DEPTH)
-			mtdp->filter.depth = tr->depth;
-
+		if (tr->flags & TRIGGER_FL_DEPTH) {
+			mtdp->filter.depth = 0;
+			mtdp->filter.max_depth = max_depth = tr->depth;
+		}
 		if (tr->flags & TRIGGER_FL_TRACE_ON)
 			mcount_enabled = true;
-
 		if (tr->flags & TRIGGER_FL_TRACE_OFF)
 			mcount_enabled = false;
 
@@ -918,10 +934,10 @@ enum filter_result mcount_entry_filter_check(struct mcount_thread_data *mtdp, un
 
 #undef FLAGS_TO_CHECK
 
-	if (mtdp->filter.depth == 0)
+	if (mtdp->filter.depth >= max_depth)
 		return FILTER_OUT;
 
-	mtdp->filter.depth--;
+	mtdp->filter.depth++;
 	return FILTER_IN;
 }
 
@@ -996,19 +1012,31 @@ skip:
 	symbol_putname(sym, symname);
 }
 
-/* save current filter state to rstack */
+/**
+ * filter_save_to_rstack - save current filter state to rstack
+ * @mtdp - thread data
+ *
+ * The current values can be overwritten by triggers, and will be restored from
+ * @rstack at function exit.
+ */
+static void filter_save_to_rstack(struct mcount_thread_data *mtdp, struct mcount_ret_stack *rstack)
+{
+	rstack->filter_depth = mtdp->filter.saved_depth;
+	rstack->filter_max_depth = mtdp->filter.saved_max_depth;
+	rstack->filter_time = mtdp->filter.saved_time;
+	rstack->filter_size = mtdp->filter.saved_size;
+}
+
 void mcount_entry_filter_record(struct mcount_thread_data *mtdp, struct mcount_ret_stack *rstack,
 				struct uftrace_trigger *tr, struct mcount_regs *regs)
 {
 	if (mtdp->filter.out_count > 0 ||
-	    (mtdp->filter.in_count == 0 && mcount_filter_mode == FILTER_MODE_IN) ||
+	    (mtdp->filter.in_count == 0 && mcount_get_filter_mode() == FILTER_MODE_IN) ||
 	    (mtdp->filter.size > 0 &&
 	     mcount_getsize(&mcount_sym_info, rstack->child_ip) < mtdp->filter.size))
 		rstack->flags |= MCOUNT_FL_NORECORD;
 
-	rstack->filter_depth = mtdp->filter.saved_depth;
-	rstack->filter_time = mtdp->filter.saved_time;
-	rstack->filter_size = mtdp->filter.saved_size;
+	filter_save_to_rstack(mtdp, rstack);
 
 #define FLAGS_TO_CHECK                                                                             \
 	(TRIGGER_FL_FILTER | TRIGGER_FL_RETVAL | TRIGGER_FL_TRACE | TRIGGER_FL_FINISH |            \
@@ -1104,12 +1132,27 @@ void mcount_entry_filter_record(struct mcount_thread_data *mtdp, struct mcount_r
 #undef FLAGS_TO_CHECK
 }
 
-/* restore filter state from rstack */
+/**
+ * filter_restore_from_rstack - restore filters to their value at function entry
+ * @mtdp - thread data
+ */
+static void filter_restore_from_rstack(struct mcount_thread_data *mtdp,
+				       struct mcount_ret_stack *rstack)
+{
+	mtdp->filter.depth = rstack->filter_depth;
+	mtdp->filter.max_depth = rstack->filter_max_depth;
+	mtdp->filter.time = rstack->filter_time;
+	mtdp->filter.size = rstack->filter_size;
+}
+
 void mcount_exit_filter_record(struct mcount_thread_data *mtdp, struct mcount_ret_stack *rstack,
 			       long *retval)
 {
 	uint64_t time_filter = mtdp->filter.time;
 
+	if (time_filter == FILTER_NO_TIME)
+		time_filter = mcount_threshold;
+
 	pr_dbg3("<%d> exit  %lx\n", mtdp->idx, rstack->child_ip);
 
 #define FLAGS_TO_CHECK (MCOUNT_FL_FILTERED | MCOUNT_FL_NOTRACE | MCOUNT_FL_RECOVER)
@@ -1126,9 +1169,7 @@ void mcount_exit_filter_record(struct mcount_thread_data *mtdp, struct mcount_re
 
 #undef FLAGS_TO_CHECK
 
-	mtdp->filter.depth = rstack->filter_depth;
-	mtdp->filter.time = rstack->filter_time;
-	mtdp->filter.size = rstack->filter_size;
+	filter_restore_from_rstack(mtdp, rstack);
 
 	if (!(rstack->flags & MCOUNT_FL_NORECORD)) {
 		if (mtdp->record_idx > 0)
@@ -1144,7 +1185,7 @@ void mcount_exit_filter_record(struct mcount_thread_data *mtdp, struct mcount_re
 			struct uftrace_trigger tr;
 
 			/* there's a possibility of overwriting by return value */
-			uftrace_match_filter(rstack->child_ip, &mcount_triggers, &tr);
+			uftrace_match_filter(rstack->child_ip, &mcount_triggers->root, &tr);
 			save_trigger_read(mtdp, rstack, tr.read, true);
 		}
 
@@ -1152,7 +1193,7 @@ void mcount_exit_filter_record(struct mcount_thread_data *mtdp, struct mcount_re
 			save_watchpoint(mtdp, rstack, mcount_watchpoints);
 
 		if (((rstack->end_time - rstack->start_time > time_filter) &&
-		     (!mcount_has_caller || rstack->flags & MCOUNT_FL_CALLER)) ||
+		     (!mcount_triggers->caller_count || rstack->flags & MCOUNT_FL_CALLER)) ||
 		    rstack->flags & (MCOUNT_FL_WRITTEN | MCOUNT_FL_TRACE)) {
 			if (record_trace_data(mtdp, rstack, retval) < 0)
 				pr_err("error during record");
@@ -1758,6 +1799,56 @@ static void mcount_script_init(enum uftrace_pattern_type patt_type)
 	strv_free(&info.cmds);
 }
 
+/**
+ * swap_triggers - atomically swap the pointer to a filter rbtree and free the
+ * old one
+ * @old - pointer to the tree to deprecate
+ * @new - new version of the tree to use
+ */
+static void swap_triggers(struct uftrace_triggers_info **old, struct uftrace_triggers_info *new)
+{
+	struct uftrace_triggers_info *tmp;
+	tmp = __sync_val_compare_and_swap(old, *old, new);
+	sleep(1); /* RCU-like grace period */
+	uftrace_cleanup_triggers(tmp);
+	free(tmp);
+}
+
+/**
+ * agent_setup_filter - update the registered filters from the agent
+ * @filter_str - filters to add or remove
+ * @triggers   - rbtree of tracing filters
+ */
+static void agent_setup_filter(char *filter_str, struct uftrace_triggers_info *triggers)
+{
+	uftrace_setup_filter(filter_str, &mcount_sym_info, triggers, &mcount_filter_setting);
+}
+
+/**
+ * agent_setup_caller_filter - update the registered caller filters from the agent
+ * @caller_str - caller filters to add or remove
+ * @triggers   - rbtree where the filters are stored
+ */
+static void agent_setup_caller_filter(char *caller_str, struct uftrace_triggers_info *triggers)
+{
+	uftrace_setup_caller_filter(caller_str, &mcount_sym_info, triggers, &mcount_filter_setting);
+}
+
+/**
+ * agent_setup_trigger - update the registered triggers from the agent
+ * @trigger_str - trigger to add or remove
+ * @triggers    - rbtree of tracing filters
+ */
+static void agent_setup_trigger(char *trigger_str, struct uftrace_triggers_info *triggers)
+{
+	uftrace_setup_trigger(trigger_str, &mcount_sym_info, triggers, &mcount_filter_setting);
+}
+
+/**
+ * agent_init - initialize the agent
+ * @addr - client socket
+ * @return - socket file descriptor (-1 on error)
+ */
 static int agent_init(struct sockaddr_un *addr)
 {
 	int sfd;
@@ -1769,7 +1860,7 @@ static int agent_init(struct sockaddr_un *addr)
 		}
 	}
 
-	sfd = socket_create(addr, getpid());
+	sfd = agent_socket_create(addr, getpid());
 	if (sfd == -1)
 		return sfd;
 
@@ -1778,7 +1869,7 @@ static int agent_init(struct sockaddr_un *addr)
 		goto error;
 	}
 
-	if (socket_listen(sfd, addr) == -1)
+	if (agent_listen(sfd, addr) == -1)
 		goto error;
 
 	return sfd;
@@ -1788,12 +1879,130 @@ error:
 	return -1;
 }
 
+/**
+ * agent_fini - finalize the agent thread execution
+ * @addr - client socket
+ * @sfd - client socket file descriptor
+ */
 static void agent_fini(struct sockaddr_un *addr, int sfd)
 {
 	if (sfd != -1)
 		close(sfd);
 
 	socket_unlink(addr);
+
+	pr_dbg("agent terminated\n");
+}
+
+/**
+ * agent_read_option - fetch option type and value from agent socket
+ * @fd - socket file descriptor
+ * @opt - option type
+ * @value - option value
+ * @read_size - size of data to read
+ * @return - size of data read into @value
+ */
+static int agent_read_option(int fd, int *opt, void **value, size_t read_size)
+{
+	size_t opt_size = sizeof(*opt);
+	size_t value_size = read_size - opt_size;
+
+	if (read_all(fd, opt, opt_size) < 0)
+		return -1;
+
+	*value = realloc(*value, value_size);
+	if (!value)
+		return -1;
+
+	if (read_all(fd, *value, value_size) < 0)
+		return -1;
+
+	pr_dbg4("read agent option (size=%d)\n", read_size);
+	return value_size;
+}
+
+/**
+ * agent_apply_option - change libmcount parameters at runtime
+ * @opt      - option to apply
+ * @value    - value for the given option
+ * @size     - size of @value
+ * @triggers - triggers definition and counters
+ * @return   - 0 on success, -1 on failure
+ */
+static int agent_apply_option(int opt, void *value, size_t size,
+			      struct uftrace_triggers_info *triggers)
+{
+	struct uftrace_opts opts;
+	int ret = 0;
+	int trace;
+
+	switch (opt) {
+	case UFTRACE_AGENT_OPT_TRACE:
+		trace = *((int *)value);
+		if (mcount_enabled != trace) {
+			mcount_enabled = trace;
+			pr_dbg("turn trace %s\n", mcount_enabled ? "on" : "off");
+		}
+		break;
+
+	case UFTRACE_AGENT_OPT_DEPTH:
+		opts.depth = *((int *)value);
+		if (opts.depth != mcount_depth) {
+			mcount_depth = opts.depth;
+			pr_dbg3("dynamic depth: %d\n", mcount_depth);
+		}
+		else
+			pr_dbg3("dynamic depth unchanged\n");
+		break;
+
+	case UFTRACE_AGENT_OPT_THRESHOLD:
+		opts.threshold = *((uint64_t *)value);
+		if (opts.threshold != mcount_threshold) {
+			mcount_threshold = opts.threshold;
+			pr_dbg3("dynamic time threshold: %lu\n", mcount_threshold);
+		}
+		else
+			pr_dbg3("dynamic time threshold unchanged\n");
+		break;
+
+	case UFTRACE_AGENT_OPT_PATTERN:
+		opts.patt_type = *((int *)value);
+		if (opts.patt_type != mcount_filter_setting.ptype) {
+			mcount_filter_setting.ptype = opts.patt_type;
+			pr_dbg3("use pattern type %#x\n", opts.patt_type);
+		}
+		break;
+
+	case UFTRACE_AGENT_OPT_FILTER:
+		pr_dbg3("apply filter '%s' (size=%d)\n", value, size);
+		agent_setup_filter(value, triggers);
+		break;
+
+	case UFTRACE_AGENT_OPT_CALLER:
+		pr_dbg3("apply caller filter '%s' (size=%d)\n", value, size);
+		agent_setup_caller_filter(value, triggers);
+		break;
+
+	case UFTRACE_AGENT_OPT_TRIGGER:
+		pr_dbg3("apply trigger '%s' (size=%d)\n", value, size);
+		agent_setup_trigger(value, triggers);
+		break;
+
+	default:
+		ret = -1;
+	}
+
+	return ret;
+}
+
+static bool triggers_needs_copy(int opt)
+{
+	bool ret;
+#define MATCHING_OPTIONS                                                                           \
+	(UFTRACE_AGENT_OPT_FILTER | UFTRACE_AGENT_OPT_CALLER | UFTRACE_AGENT_OPT_TRIGGER)
+	ret = opt & MATCHING_OPTIONS;
+#undef MATCHING_OPTIONS
+	return ret;
 }
 
 /* Agent routine, applying instructions from the CLI. */
@@ -1801,56 +2010,110 @@ void *agent_apply_commands(void *arg)
 {
 	int sfd, cfd; /* socket fd, connection fd */
 	bool close_connection;
-	enum uftrace_dopt dopt;
+	struct uftrace_msg msg;
 	struct sockaddr_un addr;
+	void *value = NULL;
+	size_t size;
+	struct uftrace_triggers_info *triggers_copy = NULL;
 
+	/* initialize agent */
 	sfd = agent_init(&addr);
 	if (sfd == -1) {
 		pr_warn("agent cannot start\n");
 		return NULL;
 	}
 	agent_run = true;
-	pr_dbg("agent started on socket %s\n", addr.sun_path);
+	pr_dbg("agent started on socket '%s'\n", addr.sun_path);
 
+	/* handle incoming connections consecutively */
 	while (agent_run) {
-		cfd = socket_accept(sfd);
+		cfd = agent_accept(sfd);
 		if (cfd == -1) {
-			pr_warn("error accepting socket connection\n");
+			pr_dbg2("error accepting socket connection\n");
 			continue;
 		}
-		pr_dbg2("agent connection open\n");
+		pr_dbg3("client connected\n");
 
+		/* read client messages */
 		close_connection = false;
 		while (!close_connection) {
-			if (read(cfd, &dopt, sizeof(enum uftrace_dopt)) == -1) {
-				pr_warn("error reading option\n", errno);
-				break;
+			int status = 0;
+			int opt;
+
+			/* read message header to get type */
+			if (agent_message_read_head(cfd, &msg) == -1) {
+				status = EINVAL;
+				pr_dbg3("error reading client message\n");
+				agent_message_send(cfd, UFTRACE_MSG_AGENT_ERR, &status,
+						   sizeof(status));
+				continue;
 			}
 
-			switch (dopt) {
-			case UFTRACE_DOPT_CLOSE:
+			/* parse message body */
+			switch (msg.type) {
+			case UFTRACE_MSG_AGENT_QUERY:
+				status = MCOUNT_AGENT_CAPABILITIES;
+				pr_dbg3("send capabilities to client\n");
+				agent_message_send(cfd, UFTRACE_MSG_AGENT_OK, &status,
+						   sizeof(status));
+				break;
+
+			case UFTRACE_MSG_AGENT_SET_OPT:
+				size = agent_read_option(cfd, &opt, &value, msg.len);
+				if (status < 0) {
+					status = EINVAL;
+					agent_message_send(cfd, UFTRACE_MSG_AGENT_ERR, &status,
+							   sizeof(status));
+					break;
+				}
+
+				/* deep copy mcount_triggers for each connection (if needed) */
+				if (triggers_needs_copy(opt) && !triggers_copy) {
+					triggers_copy = xmalloc(sizeof(*triggers_copy));
+					*triggers_copy =
+						uftrace_deep_copy_triggers(mcount_triggers);
+				}
+				status = agent_apply_option(opt, value, size, triggers_copy);
+				if (status == 0)
+					agent_message_send(cfd, UFTRACE_MSG_AGENT_OK, NULL, 0);
+				else
+					agent_message_send(cfd, UFTRACE_MSG_AGENT_ERR, &status,
+							   sizeof(status));
+				break;
+
+			case UFTRACE_MSG_AGENT_GET_OPT:
+				/* TODO send data */
+				agent_message_send(cfd, UFTRACE_MSG_AGENT_OK, NULL, 0);
+				break;
+
+			case UFTRACE_MSG_AGENT_CLOSE:
 				close_connection = true;
-				if (agent_run)
-					socket_send_option(cfd, UFTRACE_DOPT_CLOSE, NULL, 0);
+				agent_message_send(cfd, UFTRACE_MSG_AGENT_OK, NULL, 0);
 				break;
 
 			default:
 				close_connection = true;
-				pr_warn("option not recognized: %d\n", dopt);
+				pr_dbg3("agent message not recognized\n");
 			}
 		}
+
+		if (triggers_copy) {
+			swap_triggers(&mcount_triggers, triggers_copy);
+			triggers_copy = NULL;
+		}
+
 		if (close(cfd) == -1)
-			pr_dbg2("cannot close socket connection\n");
+			pr_dbg3("error closing client socket\n");
 		else
-			pr_dbg2("agent connection closed\n");
+			pr_dbg3("client disconnected\n");
 	}
 
+	free(value);
 	agent_fini(&addr, sfd);
-	pr_dbg("agent exited\n");
 	return 0;
 }
 
-static void agent_spawn()
+static void agent_spawn(void)
 {
 	errno = pthread_create(&agent, NULL, &agent_apply_commands, NULL);
 	if (errno != 0)
@@ -1859,38 +2122,44 @@ static void agent_spawn()
 
 /* Check if the agent is up. If so, set its run flag to false, open and close
  * connection . */
-static void agent_kill()
+static void agent_kill(void)
 {
 	int sfd;
+	int status;
 	struct sockaddr_un addr;
+	struct uftrace_msg ack;
 
 	if (!agent_run)
 		return;
-
 	agent_run = false;
 
-	sfd = socket_create(&addr, getpid());
+	sfd = agent_socket_create(&addr, getpid());
 	if (sfd == -1)
-		return; /* Agent must have already exited */
+		goto error;
 
-	if (socket_connect(sfd, &addr) == -1) {
+	if (agent_connect(sfd, &addr) == -1) {
 		if (errno != ENOENT) /* The agent may have ended and deleted the socket */
 			goto error;
 	}
 
-	if (socket_send_option(sfd, UFTRACE_DOPT_CLOSE, NULL, 0) == -1) {
-		pr_dbg("cannot stop agent loop\n");
+	status = agent_message_send(sfd, UFTRACE_MSG_AGENT_CLOSE, NULL, 0);
+	if (status < 0)
+		goto error;
+	status = agent_message_read_response(sfd, &ack);
+	if (status < 0 || ack.type != UFTRACE_MSG_AGENT_OK)
 		goto error;
-	}
 
 	close(sfd);
 
 	if (pthread_join(agent, NULL) != 0)
 		pr_dbg("agent left in unknown state\n");
+
 	return;
 
 error:
-	agent_fini(&addr, sfd);
+	pr_dbg2("error terminating agent routine\n ");
+	close(sfd);
+	socket_unlink(&addr);
 }
 
 static __used void mcount_startup(void)
@@ -1913,7 +2182,6 @@ static __used void mcount_startup(void)
 	char *symdir_str;
 	struct stat statbuf;
 	bool nest_libcall;
-	enum uftrace_pattern_type patt_type = PATT_REGEX;
 
 	if (!(mcount_global_flags & MCOUNT_GFL_SETUP))
 		return;
@@ -2014,14 +2282,14 @@ static __used void mcount_startup(void)
 	record_proc_maps(dirname, mcount_session_name(), &mcount_sym_info);
 
 	if (pattern_str)
-		patt_type = parse_filter_pattern(pattern_str);
+		mcount_filter_setting.ptype = parse_filter_pattern(pattern_str);
 
 	if (patch_str)
 		mcount_return_fn = (unsigned long)dynamic_return;
 	else
 		mcount_return_fn = (unsigned long)mcount_return;
 
-	mcount_filter_init(patt_type, !!patch_str);
+	mcount_filter_init(&mcount_filter_setting, !!patch_str);
 	mcount_watch_init();
 
 	if (maxstack_str)
@@ -2034,10 +2302,10 @@ static __used void mcount_startup(void)
 		mcount_min_size = strtoul(minsize_str, NULL, 0);
 
 	if (patch_str)
-		mcount_dynamic_update(&mcount_sym_info, patch_str, patt_type);
+		mcount_dynamic_update(&mcount_sym_info, patch_str, mcount_filter_setting.ptype);
 
 	if (event_str)
-		mcount_setup_events(dirname, event_str, patt_type);
+		mcount_setup_events(dirname, event_str, mcount_filter_setting.ptype);
 
 	if (getenv("UFTRACE_KERNEL_PID_UPDATE"))
 		kernel_pid_update = true;
@@ -2062,7 +2330,7 @@ static __used void mcount_startup(void)
 
 	/* initialize script binding */
 	if (SCRIPT_ENABLED && script_str)
-		mcount_script_init(patt_type);
+		mcount_script_init(mcount_filter_setting.ptype);
 
 	compiler_barrier();
 	pr_dbg("mcount setup done\n");
@@ -2147,6 +2415,21 @@ void __visible_default __cyg_profile_func_exit(void *child, void *parent)
 }
 UFTRACE_ALIAS(__cyg_profile_func_exit);
 
+bool mcount_is_main_executable(const char *filename, const char *exename)
+{
+	/* on Linux main executable has empty name
+	   whereas on Android we need to compare with exename */
+	char filename_canonized[PATH_MAX];
+	char exename_canonized[PATH_MAX];
+
+	if (!*filename)
+		return true;
+	if (realpath(filename, filename_canonized) && realpath(exename, exename_canonized)) {
+		return strcmp(filename_canonized, exename_canonized) == 0;
+	}
+	return false;
+}
+
 #ifndef UNIT_TEST
 /*
  * Initializer and Finalizer
@@ -2171,6 +2454,10 @@ static void setup_mcount_test(void)
 	mcount_exename = read_exename();
 	pthread_key_create(&mtd_key, mtd_dtor);
 	mcount_global_flags = 0;
+
+	mcount_triggers = xmalloc(sizeof(*mcount_triggers));
+	memset(mcount_triggers, 0, sizeof(*mcount_triggers));
+	mcount_triggers->root = RB_ROOT;
 }
 
 #define SHMEM_SESSION_FMT "/uftrace-%s-%d-%03d"
