@@ -468,6 +468,7 @@ static int setup_mod_plthook_data(struct dl_phdr_info *info, size_t sz, void *ar
 		"libmcount-fast.so",
 		"libmcount-single.so",
 		"libmcount-fast-single.so",
+		"uftrace_python.so",
 		/* system base libraries */
 		"libc.so.6",
 		"libc-2.*.so",
@@ -481,6 +482,11 @@ static int setup_mod_plthook_data(struct dl_phdr_info *info, size_t sz, void *ar
 		"ld-linux-*.so.*",
 		"libdl.so.2",
 		"libdl-2.*.so",
+#ifdef __ANDROID__
+		"linker64",
+		"libc.so",
+		"libm.so",
+#endif
 	};
 	size_t k;
 	static bool exe_once = true;
@@ -511,6 +517,9 @@ static int setup_exe_plthook_data(struct dl_phdr_info *info, size_t sz, void *ar
 	const char *exename = arg;
 	unsigned long offset = info->dlpi_addr;
 
+	if (!mcount_is_main_executable(info->dlpi_name, exename))
+		return 0;
+
 	pr_dbg2("setup plthook data for %s (offset: %lx)\n", exename, offset);
 
 	hook_pltgot(exename, offset);
