@@ -86,6 +86,7 @@ class TestBase:
     origdir = os.getcwd()
     basedir = os.path.dirname(origdir)
     objdir = 'objdir' in os.environ and os.environ['objdir'] or basedir
+    srcdir = 'srcdir' in os.environ and os.environ['srcdir'] or basedir
     uftrace_cmd = objdir + '/uftrace'
     default_opt = '--no-pager --no-event --libmcount-path=' + objdir
 
@@ -108,6 +109,8 @@ class TestBase:
         self.subcmd = 'live'
         self.option = ''
         self.exearg = 't-' + name
+        self.p_flag = ''
+        self.p_libs = []
         self.test_feature()
         if Elf.get_elf_machine(self.uftrace_cmd) == 'i386':
             self.default_cflags.append('-m32')
@@ -154,9 +157,16 @@ class TestBase:
         cmd[src_idx] = abs_src
         return " ".join(cmd)
 
+    def strip_tracing_flags(self, cmd):
+        cmd = cmd.replace('-pg', '').replace('-finstrument-functions', '')
+        cmd = re.sub(r'-fpatchable-function-entry=[0-9]+', '', cmd)
+        return cmd
+
     def build_it(self, build_cmd):
         build_cmd = self.convert_abs_path(build_cmd)
 
+        if '-fpatchable-function-entry' in build_cmd:
+            self.p_flag = '-P .'
         try:
             p = sp.Popen(build_cmd.split(), stderr=sp.PIPE)
             if p.wait() != 0:
@@ -202,7 +212,7 @@ class TestBase:
         build_cmd = '%s -o lib%s.so %s s-%s.c %s' % \
                     (lang['cc'], dstname, lib_cflags, srcname, build_ldflags)
 
-        build_cmd = build_cmd.replace('-pg', '').replace('-finstrument-functions', '')
+        build_cmd = self.strip_tracing_flags(build_cmd)
         self.pr_debug("build command for library: %s" % build_cmd)
         return self.build_it(build_cmd)
 
@@ -216,6 +226,9 @@ class TestBase:
 
         lib_cflags = build_cflags + ' -shared -fPIC'
 
+        if '-fpatchable-function-entry' in cflags:
+            self.p_libs.append('libabc_test_lib.so')
+
         # build libabc_test_lib.so library
         build_cmd = '%s -o libabc_test_lib.so %s s-lib.c %s' % \
                     (lang['cc'], lib_cflags, build_ldflags)
@@ -233,6 +246,9 @@ class TestBase:
 
         lib_cflags = build_cflags + ' -shared -fPIC'
 
+        if '-fpatchable-function-entry' in cflags:
+            self.p_libs.append('lib%s.so' % name)
+
         # build lib{foo}.so library
         build_cmd = '%s -o lib%s.so %s s-lib%s%s %s' % \
                     (lang['cc'], name, lib_cflags, name, lang['ext'], build_ldflags)
@@ -251,12 +267,16 @@ class TestBase:
                                  [self.cflags, cflags, os.getenv(lang['flags'], '')])
         build_ldflags = ' '.join([self.ldflags, ldflags, os.getenv('LDFLAGS', '')])
         exe_ldflags = build_ldflags + ' -Wl,-rpath,$ORIGIN -L. '
+
+        if '-fpatchable-function-entry' in cflags:
+            self.p_libs.append(prog)
+
         for lib in libs:
             exe_ldflags += ' -l' + lib[3:-3]
 
         build_cmd = '%s -o %s %s %s %s' % (lang['cc'], prog, build_cflags, srcname, exe_ldflags)
         if not instrument:
-            build_cmd = build_cmd.replace('-pg', '').replace('-finstrument-functions', '')
+            build_cmd = self.strip_tracing_flags(build_cmd)
 
         self.pr_debug("build command for executable: %s" % build_cmd)
         return self.build_it(build_cmd)
@@ -272,8 +292,24 @@ class TestBase:
     def runcmd(self):
         """ This function returns (shell) command that runs the test.
             A test case can extend this to setup a complex configuration.  """
-        return '%s %s %s %s %s' % (TestBase.uftrace_cmd, self.subcmd, \
-                                   TestBase.default_opt, self.option, self.exearg)
+
+        if len(self.p_libs) > 0:
+            self.p_flag = ''
+            for lib in self.p_libs:
+                self.p_flag += '-P .@%s ' % lib
+
+        if '-P' in self.option:
+            self.p_flag = ''
+
+        # On aarch64, gcc and clang emit calls to memcpy() to pass and return structs.
+        # This depends on a number of factors (gcc, clang, size and optimisation levels).
+        # For now, it affects all tests tracing s-arg.c:pass() and return.c:return_large().
+        # Any test can be affected by it and it may change from compiler to compiler.
+        # To prevent inconsistencies, filter all calls to memcpy() for all tests:
+        self.option += ' -N memcpy'
+
+        return '%s %s %s %s %s %s' % (TestBase.uftrace_cmd, self.subcmd, \
+                                   TestBase.default_opt, self.p_flag, self.option, self.exearg)
 
     def task_sort(self, output, ignore_children=False):
         """ This function post-processes output of the test to be compared .
@@ -282,7 +318,7 @@ class TestBase:
         order = 1
         before_main = True
         for ln in output.split('\n'):
-            if ln.find(' | main()') > 0:
+            if ln.find(' | main()') > 0 or ln.find(' | __main__.<module>') > 0:
                 before_main = False
             if before_main:
                 continue
@@ -346,7 +382,7 @@ class TestBase:
             # [0]         [1]   [2]        [3]   [4]     [5]
             # total_time  unit  self_time  unit  called  function
             try:
-                if line[5].startswith('__'):
+                if line[-1].startswith('__'):
                     continue
             except:
                 pass
@@ -498,7 +534,7 @@ class TestBase:
             v = int(f.readline())
             f.close()
 
-            if v == 3:
+            if v >= 3:
                 return False
         except:
             pass
@@ -536,6 +572,8 @@ class TestBase:
         self.subcmd = 'live'
         self.option = ''
         self.exearg = 't-' + self.name
+        if self.lang == 'Python':
+            self.exearg = TestBase.srcdir + '/tests/' + 's-' + self.name + '.py'
 
         cmd = self.prepare()
         if cmd == '':
@@ -598,8 +636,12 @@ class TestBase:
             timed_out = True
         signal.alarm(0)
 
-        result_expect = self.sort(self.result)
-        result_tested = self.sort(result_origin)  # for python3
+        try:
+            result_tested = self.sort(result_origin)  # for python3, may fail!
+            result_expect = self.sort(self.result)
+        except IndexError:
+            result_tested = result_origin
+            result_expect = self.result
 
         # strip trailing whitespace for each line.
         result_expect = '\n'.join([line.rstrip() for line in result_expect.split('\n')])
@@ -628,8 +670,11 @@ class TestBase:
             return TestBase.TEST_DIFF_RESULT, dif
 
         if result_expect != result_tested:
-            result_expect = self.sort(self.fixup(cflags, self.result))
-            ret = TestBase.TEST_SUCCESS_FIXED
+            try:
+                result_expect = self.sort(self.fixup(cflags, self.result))
+                ret = TestBase.TEST_SUCCESS_FIXED
+            except IndexError:
+                return TestBase.TEST_DIFF_RESULT, "Internal error: Expected more results"
 
         if result_expect != result_tested:
             if diff:
@@ -639,7 +684,9 @@ class TestBase:
                 f = open('result', 'w')
                 f.write(result_tested + '\n')
                 f.close()
-                dif = "%s: diff result of %s\n" % (name, cflags)
+
+                compiler = self.supported_lang['C']['cc']
+                dif = "%s: diff result of %s %s\n" % (name, compiler, cflags)
                 try:
                     p = sp.Popen(['colordiff', '-U1', 'expect', 'result'], stdout=sp.PIPE)
                 except:
@@ -657,6 +704,17 @@ class TestBase:
         else:
             sp.call(['rm', '-rf', self.test_dir])
 
+class PyTestBase(TestBase):
+    def __init__(self, name, result, lang='Python', cflags='', ldflags='', sort='simple', serial=False):
+        TestBase.__init__(self, name, result, lang, cflags, ldflags, sort, serial)
+        # setup PYTHONPATH to load the new code before the inst
+        orig_path = os.environ["PYTHONPATH"] if "PYTHONPATH" in os.environ else ""
+        os.environ["PYTHONPATH"] = TestBase.objdir + '/python'
+        if TestBase.objdir != TestBase.srcdir:
+            os.environ["PYTHONPATH"] += ':' + TestBase.objdir + '/python'
+        if orig_path != "":
+            os.environ["PYTHONPATH"] += ':' + orig_path
+
 RED     = '\033[1;31m'
 GREEN   = '\033[1;32m'
 YELLOW  = '\033[1;33m'
@@ -707,22 +765,50 @@ def check_serial_case(case):
     return tc.serial
 
 
+def run_python_case(T, case, timeout):
+    tc = T.TestCase()
+    tc.set_debug(arg.debug)
+    tc.set_keep(arg.keep)
+
+    # to load uftrace.py and uftrace_python.so module
+    sys.path.append(TestBase.objdir + '/python')
+    sys.path.append(TestBase.srcdir + '/python')
+
+    ret = tc.prerun(timeout)
+    dif = ''
+    if ret == TestBase.TEST_SUCCESS:
+        ret, dif = tc.run(case, "", arg.diff, timeout)
+        ret = tc.postrun(ret)
+    return (ret, dif)
+
 def run_single_case(case, flags, opts, arg, compilers):
     result = []
+    timeout = int(arg.timeout)
 
     # for python3
     _locals = {}
-    exec("import %s; tc = %s.TestCase()" % (case, case), globals(), _locals)
-    tc = _locals['tc']
-    tc.set_debug(arg.debug)
-    tc.set_keep(arg.keep)
-    timeout = int(arg.timeout)
+    exec("import %s as T" % (case), globals(), _locals)
+    T = _locals['T']
 
     for compiler in compilers:
-        tc.set_compiler(compiler)
+        if compiler == 'python':
+            ret, dif = run_python_case(T, case, timeout)
+            result.append((ret, dif))
+            continue
+
         for flag in flags:
             for opt in opts:
+                tc = T.TestCase()
+                tc.set_debug(arg.debug)
+                tc.set_keep(arg.keep)
+                tc.set_compiler(compiler)
+
                 cflags = ' '.join(["-" + flag, "-" + opt])
+                # add -fno-ipa-sra to prevent function renames like foo.isra.0
+                # this is available on GCC only
+                if compiler == 'gcc':
+                    cflags += ' -fno-ipa-sra'
+
                 dif = ''
                 ret = tc.build(tc.name, cflags)
                 if ret == TestBase.TEST_SUCCESS:
@@ -789,9 +875,9 @@ def print_test_result(case, result, diffs, color, ftests, nr_compilers):
 def print_test_header(opts, flags, ftests, compilers):
     optslen = len(opts)
     header1 = '%-24s ' % 'Compiler'
-    header2 = '%-24s ' % 'Test case'
+    header2 = '%-24s ' % 'Runtime test case'
     header3 = '-' * 24 + ':'
-    empty = '                             '
+    empty = ' ' * 100
 
     for i, compiler in enumerate(compilers):
         if i != 0:
@@ -804,6 +890,7 @@ def print_test_header(opts, flags, ftests, compilers):
             header3 += ' ' + opts
         header1 += ' ' + compiler[:optslen*len(flags)+len(flags)-1] + empty[len(compiler):len(header3)-len(header1)-1]
 
+    print("")
     print(header1)
     print(header2)
     print(header3)
@@ -813,6 +900,18 @@ def print_test_header(opts, flags, ftests, compilers):
     ftests.flush()
 
 
+def print_python_test_header(ftests):
+    header1 = '%-24s  %s' % ('Python test case', 'Result')
+    header2 = '-' * 32
+
+    print("")
+    print(header1)
+    print(header2)
+    ftests.write(header1 + '\n')
+    ftests.write(header2 + '\n')
+    ftests.flush()
+
+
 def print_test_report(color, shared):
     success = shared.stats[TestBase.TEST_SUCCESS] + shared.stats[TestBase.TEST_SUCCESS_FIXED]
     percent = 100.0 * success / shared.total
@@ -834,16 +933,18 @@ def parse_argument():
 
     parser = argparse.ArgumentParser()
     parser.add_argument("-f", "--profile-flags", dest='flags',
-                        default="pg finstrument-functions",
+                        default="pg finstrument-functions fpatchable-function-entry",
                         help="comma separated list of compiler profiling flags")
     parser.add_argument("-O", "--optimize-levels", dest='opts', default="0123s",
                         help="compiler optimization levels")
-    parser.add_argument("case", nargs='?', default="all",
-                        help="test case: 'all' or test number or (partial) name")
+    parser.add_argument("cases", nargs='?', default="all",
+                        help="test cases: 'all' or test number or (partial) name")
     parser.add_argument("-p", "--profile-pg", dest='pg_flag', action='store_true',
                         help="profiling with -pg option")
     parser.add_argument("-i", "--instrument-functions", dest='if_flag', action='store_true',
                         help="profiling with -finstrument-functions option")
+    parser.add_argument("-e", "--patchable-function-entry", dest='pfe_flag', action='store_true',
+                        help="profiling with -fpatchable-function-entry option")
     parser.add_argument("-d", "--diff", dest='diff', action='store_true',
                         help="show diff result if not matched")
     parser.add_argument("-v", "--verbose", dest='debug', action='store_true',
@@ -860,6 +961,8 @@ def parse_argument():
                         help="keep the test directories with compiled binaries")
     parser.add_argument("-q", "--quiet", dest='quiet', action='store_true',
                         help="Hide normal results and print only abnormal results.")
+    parser.add_argument("-P", "--python", dest='python', action='store_true',
+                        help="Run python test cases instead")
 
     return parser.parse_args()
 
@@ -870,15 +973,24 @@ if __name__ == "__main__":
 
     arg = parse_argument()
 
-    if arg.case == 'all':
-        testcases = glob.glob('t???_*.py')
+    testcases = []
+    if arg.cases == 'all':
+        if arg.python:
+            testcases = glob.glob('p???_*.py')
+        else:
+            testcases = glob.glob('t???_*.py')
     else:
         try:
-            testcases = glob.glob('t*' + arg.case + '*.py')
+            cases = arg.cases.split('|')
+            for case in cases:
+                if arg.python:
+                    testcases.extend(glob.glob('p*' + case + '*.py'))
+                else:
+                    testcases.extend(glob.glob('t*' + case + '*.py'))
             arg.worker = min(arg.worker, len(testcases))
         finally:
             if len(testcases) == 0:
-                print("cannot find testcase for : %s" % arg.case)
+                print("cannot find testcase for : %s" % arg.cases)
                 sys.exit(0)
 
     # Use multiprocessing pool if the number of workers is greater than 1.
@@ -886,19 +998,39 @@ if __name__ == "__main__":
 
     opts = ' '.join(sorted(['O' + o for o in arg.opts]))
 
+    patch_size = {
+        'x86_64'  : 5,
+        'aarch64' : 2,
+    }
+
+    m = os.uname()[-1]  # machine
+
     if arg.pg_flag:
         flags = ['pg']
     elif arg.if_flag:
         flags = ['finstrument-functions']
+    elif arg.pfe_flag:
+        flags = ['fpatchable-function-entry']
+        if m not in patch_size:
+            print('fpatchable-function-entry not supported on current platform\n')
+            exit(-1)
     else:
         flags = arg.flags.split()
+        if m not in patch_size:
+            flags.remove('fpatchable-function-entry')
+
+    for i in range(len(flags)):
+        if flags[i] == 'fpatchable-function-entry':
+            flags[i] += '=%d' % patch_size[m]
 
     def has_compiler(compiler):
         installed = os.system('command -v %s > /dev/null' % compiler) == 0
         return installed
 
     compilers = []
-    if arg.compiler == 'all':
+    if arg.python:
+        compilers.append('python')
+    elif arg.compiler == 'all':
         for compiler in ['gcc', 'clang']:
             if has_compiler(compiler):
                 compilers.append(compiler)
@@ -949,7 +1081,11 @@ if __name__ == "__main__":
         print("Start %s tests with %d worker" % (shared.tests_count, arg.worker))
     else:
         print("Start %s tests without worker pool" % shared.tests_count)
-    print_test_header(opts, flags, ftests, compilers)
+
+    if arg.python:
+        print_python_test_header(ftests)
+    else:
+        print_test_header(opts, flags, ftests, compilers)
 
     color = True
     if arg.color == 'auto':
