@@ -58,5 +58,5 @@ class TestCase(TestBase):
    1.995 ms [15156] | } /* main */
 """)
 
-    def runcmd(self):
-        return '%s --column-view --column-offset=4 --no-merge %s' % (TestBase.uftrace_cmd, 't-' + self.name)
+    def setup(self):
+        self.option = '--column-view --column-offset=4 --no-merge'
