@@ -1,8 +1,9 @@
 #!/usr/bin/env python
 
-from runtest import TestBase
 import os
 
+from runtest import TestBase
+
 class TestCase(TestBase):
     def __init__(self):
         TestBase.__init__(self, 'fork2', serial=True, result="""
@@ -28,7 +29,7 @@ class TestCase(TestBase):
   41.010 ms [19227] | } /* main */
 """)
 
-    def pre(self):
+    def prerun(self, timeout):
         if os.geteuid() != 0:
             return TestBase.TEST_SKIP
         if os.path.exists('/.dockerenv'):
@@ -36,15 +37,10 @@ class TestCase(TestBase):
 
         return TestBase.TEST_SUCCESS
 
-    def runcmd(self):
-        uftrace = TestBase.uftrace_cmd
-        program = 't-' + self.name
-
-        argument  = '-k -F main'
-        argument += ' -F sys_open*@kernel'
-        argument += ' -F sys_close*@kernel'
-
-        return '%s %s %s' % (uftrace, argument, program)
+    def setup(self):
+        self.option  = '-k -F main '
+        self.option += '-F sys_open*@kernel '
+        self.option += '-F sys_close*@kernel'
 
     def fixup(self, cflags, result):
         uname = os.uname()
@@ -54,7 +50,7 @@ class TestCase(TestBase):
         # Linux v4.17 (x86_64) changed syscall routines
         major, minor, release = uname[2].split('.')
         if uname[0] == 'Linux' and uname[4] == 'x86_64' and \
-           int(major) >= 4 and int(minor) >= 17:
+           int(major) >= 5 or (int(major) == 4 and int(minor) >= 17):
             result = result.replace('sys_', '__x64_sys_')
 
         return result
