@@ -1,10 +1,11 @@
 #!/usr/bin/env python
 
-from runtest import TestBase
+import os.path
 import subprocess as sp
 
+from runtest import TestBase
+
 TDIR  = 'xxx'
-TDIR2 = 'xxx/uftrace.data'
 
 class TestCase(TestBase):
     def __init__(self):
@@ -21,22 +22,30 @@ class TestCase(TestBase):
    2.405 us [28141] |   } /* a */
    3.005 us [28141] | } /* main */
 """)
+
+    def prerun(self, timeout):
         self.gen_port()
 
-    recv_p = None
-
-    def pre(self):
-        recv_cmd = '%s recv -d %s --port %s' % (TestBase.uftrace_cmd, TDIR, self.port)
+        self.subcmd = 'recv'
+        self.option = '-d %s --port %s' % (TDIR, self.port)
+        self.exearg = ''
+        recv_cmd = self.runcmd()
+        self.pr_debug("prerun command: " + recv_cmd)
         self.recv_p = sp.Popen(recv_cmd.split())
 
-        record_cmd = '%s record -H %s --port %s %s' % (TestBase.uftrace_cmd, 'localhost', self.port, 't-abc')
+        self.subcmd = 'record'
+        self.option = '--host %s --port %s' % ('localhost', self.port)
+        self.exearg = 't-' + self.name
+        record_cmd = self.runcmd()
+        self.pr_debug("prerun command: " + record_cmd)
         sp.call(record_cmd.split())
         return TestBase.TEST_SUCCESS
 
-    def runcmd(self):
-        return '%s replay -d %s' % (TestBase.uftrace_cmd, TDIR2)
+    def setup(self):
+        self.subcmd = 'replay'
+        self.option = '-d ' + os.path.join(TDIR, 'uftrace.data')
+        self.exearg = ''
 
-    def post(self, ret):
+    def postrun(self, ret):
         self.recv_p.terminate()
-        sp.call(['rm', '-rf', TDIR])
         return ret
