@@ -1,8 +1,9 @@
 #!/usr/bin/env python
 
-from runtest import TestBase
 import subprocess as sp
 
+from runtest import TestBase
+
 XDIR='xxx'
 YDIR='yyy'
 
@@ -14,30 +15,35 @@ class TestCase(TestBase):
 #  [0] base: xxx   (from uftrace record -d xxx -F main tests/t-diff 0 )
 #  [1] diff: yyy   (from uftrace record -d yyy -F main tests/t-diff 1 )
 #
-    Total time    Self time        Calls   Function
-  ============   ==========   ==========   ====================
-     +1.296 ms    +0.312 us           +0   main
-     +1.292 ms    +1.292 ms           +3   usleep
-     +1.225 ms    +2.609 us           +0   foo
-   +158.450 us    +1.305 us           +0   bar
-     -0.066 us    -0.066 us           +0   atoi
+    Total time     Self time         Calls   Function
+   ===========   ===========   ===========   ====================
+     -1.495 ms     -1.495 ms            +4   usleep
+   -320.169 us     -1.968 us            +2   bar
+     -1.270 ms     -2.370 us            +1   foo
+     +0.359 us     +0.359 us            +0   atoi
+     -1.499 ms     -0.157 us            +0   main
 """)
 
-    def pre(self):
-        record_cmd = '%s record -d %s -F main %s 0' % (TestBase.uftrace_cmd, XDIR, 't-' + self.name)
+    def prerun(self, timeout):
+        self.subcmd = 'record'
+        self.option = '-d %s -F main' % XDIR
+        self.exearg = 't-' + self.name + ' 0'
+        record_cmd = self.runcmd()
+        self.pr_debug('prerun command: ' + record_cmd)
         sp.call(record_cmd.split())
-        record_cmd = '%s record -d %s -F main %s 1' % (TestBase.uftrace_cmd, YDIR, 't-' + self.name)
+
+        self.option = '-d %s -F main' % YDIR
+        self.exearg = 't-' + self.name + ' 1'
+        record_cmd = self.runcmd()
+        self.pr_debug('prerun command: ' + record_cmd)
         sp.call(record_cmd.split())
+
         return TestBase.TEST_SUCCESS
 
-    def runcmd(self):
-        uftrace = TestBase.uftrace_cmd
-        options = '--diff-policy compact,no-percent,abs'  # new default
-        return '%s report -d %s --diff %s %s' % (uftrace, XDIR, YDIR, options)
-
-    def post(self, ret):
-        sp.call(['rm', '-rf', XDIR, YDIR])
-        return ret
+    def setup(self):
+        self.subcmd = 'report'
+        self.option = '--diff-policy compact,no-percent,abs -s call'  # new default
+        self.exearg = '-d %s --diff %s' % (XDIR, YDIR)
 
     def sort(self, output):
         """ This function post-processes output of the test to be compared .
