@@ -13,10 +13,10 @@ class TestCase(TestBase):
     1.080 ms    1.813 us           1  bar
     1.078 ms    2.892 us           1  usleep
     1.075 ms    1.075 ms           1  linux:schedule
-   70.176 us   70.176 us           1  __monstartup   # ignore this
+   70.176 us   70.176 us           1  __monstartup
    37.525 us    1.137 us           2  foo
    36.388 us   36.388 us           6  loop
-    1.200 us    1.200 us           1  __cxa_atexit   # and this too
+    1.200 us    1.200 us           1  __cxa_atexit
 """, sort='report')
 
     def prerun(self, timeout):
