@@ -15,15 +15,13 @@ class TestCase(TestBase):
 """)
 
     def build(self, name, cflags='', ldflags=''):
+        if not 'dwarf' in self.feature:
+            return TestBase.TEST_SKIP
         # cygprof doesn't support arguments now
         if cflags.find('-finstrument-functions') >= 0:
             return TestBase.TEST_SKIP
 
         return TestBase.build(self, name, cflags, ldflags)
 
-    def runcmd(self):
-        uftrace = TestBase.uftrace_cmd
-        options = '-A foo@arg1/p,arg2 -A filecmp@arg1/p,arg2/p'
-        program = 't-' + self.name
-
-        return '%s %s %s' % (uftrace, options, program)
+    def setup(self):
+        self.option = '-A foo@arg1/p,arg2 -A filecmp@arg1/p,arg2/p'
