@@ -59,6 +59,7 @@ enum uftrace_short_options {
 	OPT_avg_self,
 	OPT_color,
 	OPT_disabled,
+	OPT_trace,
 	OPT_demangle,
 	OPT_dbg_domain,
 	OPT_report,
@@ -149,7 +150,7 @@ __used static const char uftrace_help[] =
 "      --diff=DATA            Report differences\n"
 "      --diff-policy=POLICY   Control diff report policy\n"
 "                             (default: 'abs,compact,no-percent')\n"
-"      --disable              Start with tracing disabled\n"
+"      --disable              Start with tracing disabled (deprecated)\n"
 "  -D, --depth=DEPTH          Trace functions within DEPTH\n"
 "  -e, --estimate-return      Use only entry record type for safety\n"
 "      --event-full           Show all events outside of function\n"
@@ -173,6 +174,7 @@ __used static const char uftrace_help[] =
 "  -K, --kernel-depth=DEPTH   Trace kernel functions within DEPTH\n"
 "      --libmcount-single     Use single thread version of libmcount\n"
 "      --list-event           List available events\n"
+"  -L, --loc-filter=LOCATION  Only trace functions in the source LOCATION\n"
 "      --logfile=FILE         Save log messages to this file\n"
 "  -l, --nest-libcall         Show nested library calls\n"
 "      --libname              Show libname name with symbol name\n"
@@ -221,6 +223,7 @@ __used static const char uftrace_help[] =
 "      --task-newline         Interleave a newline when task is changed\n"
 "      --tid=TID[,TID,...]    Only replay those tasks\n"
 "      --time                 Print time information\n"
+"      --trace=STATE          Set the recording state: on, off (default: on)\n"
 "  -T, --trigger=FUNC@act[,act,...]\n"
 "                             Trigger action on those FUNCs\n"
 "  -U, --unpatch=FUNC         Don't apply dynamic patching for FUNCs\n"
@@ -255,6 +258,7 @@ static const struct option uftrace_options[] = {
 	REQ_ARG(time-filter, 't'),
 	REQ_ARG(caller-filter, 'C'),
 	REQ_ARG(argument, 'A'),
+	REQ_ARG(trigger, 'T'),
 	REQ_ARG(retval, 'R'),
 	NO_ARG(auto-args, 'a'),
 	NO_ARG(no-args, OPT_no_args),
@@ -284,6 +288,7 @@ static const struct option uftrace_options[] = {
 	NO_ARG(avg-self, OPT_avg_self),
 	REQ_ARG(color, OPT_color),
 	NO_ARG(disable, OPT_disabled),
+	REQ_ARG(trace, OPT_trace),
 	REQ_ARG(demangle, OPT_demangle),
 	NO_ARG(record, OPT_record),
 	NO_ARG(report, OPT_report),
@@ -681,6 +686,10 @@ static int parse_option(struct uftrace_opts *opts, int key, char *arg)
 		strv_append(&default_opts, arg);
 
 		opts->threshold = parse_time(arg, 3);
+		if (opts->threshold >= OPT_THRESHOLD_MAX) {
+			pr_use("invalid time given: %lu (ignoring..)\n", opts->threshold);
+			opts->threshold = OPT_THRESHOLD_MAX - 1;
+		}
 		if (opts->range.start || opts->range.stop) {
 			pr_use("--time-range cannot be used with --time-filter\n");
 			opts->range.start = opts->range.stop = 0;
@@ -858,7 +867,17 @@ static int parse_option(struct uftrace_opts *opts, int key, char *arg)
 		break;
 
 	case OPT_disabled:
-		opts->disabled = true;
+		pr_use("'--disable' is deprecated, use --trace=off instead.\n");
+		opts->trace = TRACE_STATE_OFF;
+		break;
+
+	case OPT_trace:
+		if (!strcmp(arg, "on"))
+			opts->trace = TRACE_STATE_ON;
+		else if (!strcmp(arg, "off"))
+			opts->trace = TRACE_STATE_OFF;
+		else
+			pr_use("unknown tracing state: %s (ignoring..)\n", arg);
 		break;
 
 	case OPT_demangle:
@@ -1371,7 +1390,6 @@ int main(int argc, char *argv[])
 		.dirname = UFTRACE_DIR_NAME,
 		.libcall = true,
 		.bufsize = SHMEM_BUFFER_SIZE,
-		.depth = OPT_DEPTH_DEFAULT,
 		.max_stack = OPT_RSTACK_DEFAULT,
 		.port = UFTRACE_RECV_PORT,
 		.use_pager = true,
@@ -1473,6 +1491,11 @@ int main(int argc, char *argv[])
 	if (opts.use_pager)
 		pager = setup_pager();
 
+	if (!opts.pid) { /* Keep uninitialized values in client mode */
+		if (!opts.depth)
+			opts.depth = OPT_DEPTH_DEFAULT;
+	}
+
 	setup_color(opts.color, pager);
 	setup_signal();
 
@@ -1484,6 +1507,11 @@ int main(int argc, char *argv[])
 	if (opts.mode == UFTRACE_MODE_TUI)
 		opts.srcline = true;
 
+	if (!opts.pid) { /* Keep uninitialized values in client mode */
+		if (opts.trace == TRACE_STATE_NONE)
+			opts.trace = TRACE_STATE_ON;
+	}
+
 	/* apply 'default.opts' options for analysis commands */
 	apply_default_opts(&argc, &argv, &opts);
 
@@ -1493,6 +1521,9 @@ int main(int argc, char *argv[])
 	argc -= opts.idx;
 	argv += opts.idx;
 
+	if (!opts.libcall && opts.nest_libcall)
+		pr_err_ns("cannot use --no-libcall and --nest-libcall options together\n");
+
 	switch (opts.mode) {
 	case UFTRACE_MODE_RECORD:
 		ret = command_record(argc, argv, &opts);
