@@ -26,6 +26,7 @@
 #define OPT_RSTACK_MAX 65535
 #define OPT_RSTACK_DEFAULT 1024
 #define OPT_DEPTH_MAX OPT_RSTACK_MAX
+#define OPT_THRESHOLD_MAX 0xFFFFFFFFFFFFFFFF /* max uint64 = 2^64-1 */
 #define OPT_DEPTH_DEFAULT OPT_RSTACK_DEFAULT
 #define OPT_COLUMN_OFFSET 8
 #define OPT_SORT_COLUMN 2
@@ -61,6 +62,7 @@ enum uftrace_feat_bits {
 	AUTO_ARGS_BIT,
 	DEBUG_INFO_BIT,
 	ESTIMATE_RETURN_BIT,
+	SYM_SIZE_BIT,
 
 	FEAT_BIT_MAX,
 
@@ -77,6 +79,7 @@ enum uftrace_feat_bits {
 	AUTO_ARGS = (1U << AUTO_ARGS_BIT),
 	DEBUG_INFO = (1U << DEBUG_INFO_BIT),
 	ESTIMATE_RETURN = (1U << ESTIMATE_RETURN_BIT),
+	SYM_SIZE = (1U << SYM_SIZE_BIT),
 };
 
 enum uftrace_info_bits {
@@ -277,7 +280,6 @@ struct uftrace_opts {
 	bool use_pager;
 	bool avg_total;
 	bool avg_self;
-	bool disabled;
 	bool report;
 	bool column_view;
 	bool want_bind_not;
@@ -308,6 +310,7 @@ struct uftrace_opts {
 	bool agent;
 	struct uftrace_time_range range;
 	enum uftrace_pattern_type patt_type;
+	enum uftrace_trace_state trace;
 };
 
 extern struct strv default_opts;
@@ -415,17 +418,19 @@ enum uftrace_msg_type {
 	UFTRACE_MSG_SEND_INFO,
 	UFTRACE_MSG_SEND_META_DATA,
 	UFTRACE_MSG_SEND_END,
-};
 
-/* Dynamic options sent by the client to the agent */
-enum uftrace_dopt {
-	UFTRACE_DOPT_CLOSE, /* Close the connection with the client */
+	UFTRACE_MSG_AGENT_CLOSE = 200, /* close the connection */
+	UFTRACE_MSG_AGENT_QUERY, /* perform connection handshake */
+	UFTRACE_MSG_AGENT_GET_OPT, /* get current option value */
+	UFTRACE_MSG_AGENT_SET_OPT, /* set new option value */
+	UFTRACE_MSG_AGENT_OK, /* ack previous message */
+	UFTRACE_MSG_AGENT_ERR, /* signal error on previous message */
 };
 
 /* msg format for communicating by pipe */
 struct uftrace_msg {
 	unsigned short magic; /* UFTRACE_MSG_MAGIC */
-	unsigned short type; /* UFTRACE_MSG_REC_* */
+	unsigned short type; /* UFTRACE_MSG_* */
 	unsigned int len;
 	unsigned char data[];
 };
@@ -453,6 +458,16 @@ struct uftrace_msg_dlopen {
 	char exename[];
 };
 
+enum uftrace_agent_opt {
+	UFTRACE_AGENT_OPT_TRACE = (1U << 0), /* turn tracing on/off */
+	UFTRACE_AGENT_OPT_DEPTH = (1U << 1), /* mcount depth filter */
+	UFTRACE_AGENT_OPT_THRESHOLD = (1U << 2), /* mcount time filter */
+	UFTRACE_AGENT_OPT_PATTERN = (1U << 3), /* pattern match type */
+	UFTRACE_AGENT_OPT_FILTER = (1U << 4), /* tracing filters */
+	UFTRACE_AGENT_OPT_CALLER = (1U << 5), /* tracing caller filters */
+	UFTRACE_AGENT_OPT_TRIGGER = (1U << 6), /* tracing trigger actions */
+};
+
 extern struct uftrace_session *first_session;
 
 void create_session(struct uftrace_session_link *sess, struct uftrace_msg_sess *msg, char *dirname,
