@@ -21,8 +21,19 @@
 bool fstack_enabled = true;
 bool live_disabled = false;
 
-static enum filter_mode fstack_filter_mode = FILTER_MODE_NONE;
-static enum filter_mode fstack_loc_mode = FILTER_MODE_NONE;
+static struct uftrace_triggers_info fstack_triggers;
+
+static inline int fstack_get_filter_mode(void)
+{
+	int filter_count = fstack_triggers.filter_count;
+	return filter_count > 0 ? FILTER_MODE_IN : FILTER_MODE_OUT;
+}
+
+static inline int fstack_get_loc_mode(void)
+{
+	int loc_count = fstack_triggers.loc_count;
+	return loc_count > 0 ? FILTER_MODE_IN : FILTER_MODE_OUT;
+}
 
 static int __read_task_ustack(struct uftrace_task_reader *task);
 
@@ -50,7 +61,7 @@ static void setup_task_handle(struct uftrace_data *handle, struct uftrace_task_r
 	task->event_color = DEFAULT_EVENT_COLOR;
 
 	/*
-	 * set display depth to non-zero only when trace-on trigger (with --disabled
+	 * set display depth to non-zero only when trace-on trigger (with --trace=off
 	 * option) or time range is set.
 	 */
 	task->display_depth_set = (fstack_enabled && !live_disabled && !handle->time_range.start);
@@ -218,8 +229,9 @@ static int setup_filters(struct uftrace_session *s, void *arg)
 {
 	struct uftrace_filter_setting *setting = arg;
 
-	uftrace_setup_filter(setting->info_str, &s->sym_info, &s->filters, &fstack_filter_mode,
-			     setting);
+	fstack_triggers.root = s->filters;
+	uftrace_setup_filter(setting->info_str, &s->sym_info, &fstack_triggers, setting);
+	s->filters = fstack_triggers.root;
 	return 0;
 }
 
@@ -227,8 +239,9 @@ static int setup_trigger(struct uftrace_session *s, void *arg)
 {
 	struct uftrace_filter_setting *setting = arg;
 
-	uftrace_setup_trigger(setting->info_str, &s->sym_info, &s->filters, &fstack_filter_mode,
-			      setting);
+	fstack_triggers.root = s->filters;
+	uftrace_setup_trigger(setting->info_str, &s->sym_info, &fstack_triggers, setting);
+	s->filters = fstack_triggers.root;
 	return 0;
 }
 
@@ -236,7 +249,9 @@ static int setup_callers(struct uftrace_session *s, void *arg)
 {
 	struct uftrace_filter_setting *setting = arg;
 
-	uftrace_setup_caller_filter(setting->info_str, &s->sym_info, &s->filters, setting);
+	fstack_triggers.root = s->filters;
+	uftrace_setup_caller_filter(setting->info_str, &s->sym_info, &fstack_triggers, setting);
+	s->filters = fstack_triggers.root;
 	return 0;
 }
 
@@ -244,7 +259,9 @@ static int setup_hides(struct uftrace_session *s, void *arg)
 {
 	struct uftrace_filter_setting *setting = arg;
 
-	uftrace_setup_hide_filter(setting->info_str, &s->sym_info, &s->filters, setting);
+	fstack_triggers.root = s->filters;
+	uftrace_setup_hide_filter(setting->info_str, &s->sym_info, &fstack_triggers, setting);
+	s->filters = fstack_triggers.root;
 	return 0;
 }
 
@@ -252,8 +269,9 @@ static int setup_locs(struct uftrace_session *s, void *arg)
 {
 	struct uftrace_filter_setting *setting = arg;
 
-	uftrace_setup_loc_filter(setting->info_str, &s->sym_info, &s->filters, &fstack_loc_mode,
-				 setting);
+	fstack_triggers.root = s->filters;
+	uftrace_setup_loc_filter(setting->info_str, &s->sym_info, &fstack_triggers, setting);
+	s->filters = fstack_triggers.root;
 	return 0;
 }
 
@@ -339,29 +357,26 @@ static int setup_fstack_filters(struct uftrace_data *handle, char *filter_str, c
 
 		setting->info_str = caller_str;
 		walk_sessions(sessions, setup_callers, setting);
-		walk_sessions(sessions, count_filters, &count);
+		walk_sessions(sessions, count_callers, &count);
 
 		if (prev == count)
 			return -1;
-	}
 
-	/* there might be caller triggers, count it separately */
-	count = 0;
-	walk_sessions(sessions, count_callers, &count);
-	if (count != 0) {
 		handle->caller_filter = true;
-		pr_dbg("setup caller filters for %d function(s)\n", count);
+		pr_dbg("setup caller filters for %d function(s)\n", count - prev);
 	}
 
 	if (hide_str) {
+		int prev = count;
+
 		setting->info_str = hide_str;
 		walk_sessions(sessions, setup_hides, setting);
 		walk_sessions(sessions, count_hides, &count);
 
-		if (count == 0)
+		if (prev == count)
 			return -1;
 
-		pr_dbg("setup hide filters for %d function(s)\n", count);
+		pr_dbg("setup hide filters for %d function(s)\n", count - prev);
 	}
 
 	if (loc_str) {
@@ -374,7 +389,7 @@ static int setup_fstack_filters(struct uftrace_data *handle, char *filter_str, c
 		if (prev == count)
 			return -1;
 
-		pr_dbg("setup location filters for %d function(s)\n", count);
+		pr_dbg("setup location filters for %d function(s)\n", count - prev);
 	}
 
 	return 0;
@@ -383,7 +398,7 @@ static int setup_fstack_filters(struct uftrace_data *handle, char *filter_str, c
 static const char *fixup_syms[] = {
 	"execl",      "execlp",	       "execle",  "execv",     "execve",      "execvp",
 	"execvpe",    "setjmp",	       "_setjmp", "sigsetjmp", "__sigsetjmp", "longjmp",
-	"siglongjmp", "__longjmp_chk", "fork",	  "vfork",     "daemon",
+	"siglongjmp", "__longjmp_chk", "fork",	  "vfork",     "daemon",      "posix.fork",
 };
 
 static int setjmp_depth;
@@ -396,13 +411,16 @@ static int build_fixup_filter(struct uftrace_session *s, void *arg)
 		.ptype = PATT_SIMPLE,
 		.auto_args = false,
 	};
+	struct uftrace_triggers_info fixups = {
+		.root = s->fixups,
+	};
 
 	pr_dbg("fixup for some special functions\n");
 
 	for (i = 0; i < ARRAY_SIZE(fixup_syms); i++) {
-		uftrace_setup_trigger((char *)fixup_syms[i], &s->sym_info, &s->fixups, NULL,
-				      &setting);
+		uftrace_setup_trigger((char *)fixup_syms[i], &s->sym_info, &fixups, &setting);
 	}
+	s->fixups = fixups.root;
 	return 0;
 }
 
@@ -421,9 +439,14 @@ static void fstack_prepare_fixup(struct uftrace_data *handle)
 static int build_arg_spec(struct uftrace_session *s, void *arg)
 {
 	struct uftrace_filter_setting *setting = arg;
+	struct uftrace_triggers_info triggers = {
+		.root = s->filters,
+	};
 
-	if (setting->info_str)
-		uftrace_setup_argument(setting->info_str, &s->sym_info, &s->filters, setting);
+	if (setting->info_str) {
+		uftrace_setup_argument(setting->info_str, &s->sym_info, &triggers, setting);
+		s->filters = triggers.root;
+	}
 
 	return 0;
 }
@@ -431,9 +454,14 @@ static int build_arg_spec(struct uftrace_session *s, void *arg)
 static int build_ret_spec(struct uftrace_session *s, void *arg)
 {
 	struct uftrace_filter_setting *setting = arg;
+	struct uftrace_triggers_info triggers = {
+		.root = s->filters,
+	};
 
-	if (setting->info_str)
-		uftrace_setup_retval(setting->info_str, &s->sym_info, &s->filters, setting);
+	if (setting->info_str) {
+		uftrace_setup_retval(setting->info_str, &s->sym_info, &triggers, setting);
+		s->filters = triggers.root;
+	}
 
 	return 0;
 }
@@ -515,7 +543,7 @@ int fstack_setup_filters(struct uftrace_opts *opts, struct uftrace_data *handle)
 		}
 	}
 
-	if (opts->disabled)
+	if (opts->trace == TRACE_STATE_OFF)
 		fstack_enabled = false;
 
 	fstack_setup_task(opts->tid, handle);
@@ -618,7 +646,8 @@ int fstack_entry(struct uftrace_task_reader *task, struct uftrace_record *rstack
 			else if (strstr(fixup->name, "longjmp")) {
 				fstack->flags |= FSTACK_FL_LONGJMP;
 			}
-			else if (strstr(fixup->name, "fork") || !strcmp(fixup->name, "daemon")) {
+			else if (strstr(fixup->name, "fork") || !strcmp(fixup->name, "daemon") ||
+				 !strcmp(fixup->name, "posix.fork")) {
 				task->fork_display_depth = task->display_depth + 1;
 			}
 		}
@@ -641,7 +670,7 @@ int fstack_entry(struct uftrace_task_reader *task, struct uftrace_record *rstack
 		task->filter.depth = task->h->depth;
 	}
 	else {
-		if (fstack_filter_mode == FILTER_MODE_IN && task->filter.in_count == 0) {
+		if (fstack_get_filter_mode() == FILTER_MODE_IN && task->filter.in_count == 0) {
 			fstack->flags |= FSTACK_FL_NORECORD;
 			return -1;
 		}
@@ -654,7 +683,7 @@ int fstack_entry(struct uftrace_task_reader *task, struct uftrace_record *rstack
 		}
 	}
 	else {
-		if (fstack_loc_mode == FILTER_MODE_IN) {
+		if (fstack_get_loc_mode() == FILTER_MODE_IN) {
 			fstack->flags |= FSTACK_FL_NORECORD;
 			return -1;
 		}
@@ -837,7 +866,8 @@ static int fstack_check_skip(struct uftrace_task_reader *task, struct uftrace_re
 		if (tr.fmode == FILTER_MODE_OUT)
 			return -1;
 	}
-	else if ((fstack_filter_mode == FILTER_MODE_IN || fstack_loc_mode == FILTER_MODE_IN) &&
+	else if ((fstack_get_filter_mode() == FILTER_MODE_IN ||
+		  fstack_get_loc_mode() == FILTER_MODE_IN) &&
 		 task->filter.in_count == 0) {
 		return -1;
 	}
@@ -974,7 +1004,7 @@ bool fstack_check_filter(struct uftrace_task_reader *task)
 	else if (task->rstack->type == UFTRACE_EVENT) {
 		/* don't change filter state, just check it */
 		if (task->filter.out_count > 0 || task->filter.depth <= 0 ||
-		    (fstack_filter_mode == FILTER_MODE_IN && task->filter.in_count == 0))
+		    (fstack_get_filter_mode() == FILTER_MODE_IN && task->filter.in_count == 0))
 			return false;
 
 		if (task->rstack->addr == EVENT_ID_PERF_SCHED_IN ||
