@@ -6,6 +6,7 @@
  * Released under the GPL v2.
  */
 
+#include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <inttypes.h>
@@ -172,17 +173,16 @@ int check_static_binary(const char *filename)
 	return ret;
 }
 
-/* caller should free the result */
-char *check_script_file(const char *filename)
+bool check_script_file(const char *filename, char *buf, size_t len)
 {
-	char *shebang = NULL;
 	char magic[2];
 	int fd;
 	char *p;
+	bool ret = false;
 
 	fd = open(filename, O_RDONLY);
 	if (fd < 0)
-		return NULL;
+		return false;
 
 	if (read(fd, magic, sizeof(magic)) < 0)
 		goto out;
@@ -190,21 +190,19 @@ char *check_script_file(const char *filename)
 	if (magic[0] != '#' || magic[1] != '!')
 		goto out;
 
-	shebang = xmalloc(1024);
-	if (read(fd, shebang, 1024) < 0) {
-		free(shebang);
-		shebang = NULL;
+	if (read(fd, buf, len) < 0) {
 		goto out;
 	}
-	shebang[1023] = '\0';
+	buf[len - 1] = '\0';
 
-	p = strchr(shebang, '\n');
+	p = strchr(buf, '\n');
 	if (p)
 		*p = '\0';
 
+	ret = true;
 out:
 	close(fd);
-	return shebang;
+	return ret;
 }
 
 static bool is_symbol_end(const char *name)
@@ -405,6 +403,9 @@ static int load_symtab(struct uftrace_symtab *symtab, const char *filename,
 		pr_dbg4("no symtab, using dynsyms instead\n");
 	}
 
+	if (iter.shdr.sh_size == 0 || iter.shdr.sh_entsize == 0)
+		goto out;
+
 	/* pre-allocate enough symbol table entries */
 	symtab->nr_alloc = iter.shdr.sh_size / iter.shdr.sh_entsize;
 	symtab->sym = xmalloc(symtab->nr_alloc * sizeof(*symtab->sym));
@@ -424,8 +425,11 @@ static int load_symtab(struct uftrace_symtab *symtab, const char *filename,
 	}
 	pr_dbg4("loaded %zd symbols\n", symtab->nr_sym);
 
-	if (symtab->nr_sym == 0)
+	if (symtab->nr_sym == 0) {
+		free(symtab->sym);
+		symtab->sym = NULL;
 		goto out;
+	}
 
 	/* also fixup the size of symbol table */
 	sort_symtab(symtab);
@@ -787,6 +791,38 @@ out:
 	return ret;
 }
 
+static void load_python_symtab(struct uftrace_sym_info *sinfo)
+{
+	char *symfile = NULL;
+	struct uftrace_mmap *map;
+
+	/* try to load python symtab (if exists) */
+	xasprintf(&symfile, "%s/%s.sym", sinfo->dirname, UFTRACE_PYTHON_SYMTAB_NAME);
+	if (access(symfile, R_OK) < 0) {
+		free(symfile);
+		return;
+	}
+
+	/* add a fake map for python script */
+	map = xzalloc(sizeof(*map) + sizeof(UFTRACE_PYTHON_SYMTAB_NAME));
+
+	memcpy(map->prot, "rwxp", 4);
+	strcpy(map->libname, UFTRACE_PYTHON_SYMTAB_NAME);
+	map->len = sizeof(UFTRACE_PYTHON_SYMTAB_NAME) - 1;
+
+	map->mod = load_module_symtab(sinfo, UFTRACE_PYTHON_SYMTAB_NAME, "no-buildid");
+	map->start = 0;
+	map->end = ALIGN(map->mod->symtab.nr_sym, 4096);
+
+	setup_debug_info(symfile, &map->mod->dinfo, 0, false);
+
+	/* add new map to symtabs */
+	map->next = sinfo->maps;
+	sinfo->maps = map;
+
+	free(symfile);
+}
+
 enum uftrace_trace_type check_trace_functions(const char *filename)
 {
 	struct uftrace_elf_data elf;
@@ -824,6 +860,10 @@ enum uftrace_trace_type check_trace_functions(const char *filename)
 
 		/* undefined function is ok here */
 		if (elf_symbol_type(&iter.sym) != STT_FUNC &&
+#ifdef __ANDROID__
+		    // Profiling functions are undefined on Android
+		    elf_symbol_type(&iter.sym) != STT_NOTYPE &&
+#endif
 		    elf_symbol_type(&iter.sym) != STT_GNU_IFUNC)
 			continue;
 
@@ -887,6 +927,7 @@ static int load_module_symbol_file(struct uftrace_symtab *symtab, const char *sy
 	while (getline(&line, &len, fp) > 0) {
 		struct uftrace_symbol *sym;
 		uint64_t addr;
+		uint32_t size;
 		char type;
 		char *name;
 		char *pos;
@@ -907,6 +948,7 @@ static int load_module_symbol_file(struct uftrace_symtab *symtab, const char *sy
 			*pos = '\0';
 
 		addr = strtoull(line, &pos, 16);
+		size = 0;
 
 		if (*pos++ != ' ') {
 			pr_dbg4("invalid symbol file format before type\n");
@@ -914,6 +956,17 @@ static int load_module_symbol_file(struct uftrace_symtab *symtab, const char *sy
 		}
 		type = *pos++;
 
+		if (isdigit(type)) {
+			/* new symbol file has size info */
+			size = strtoul(pos - 1, &pos, 16);
+
+			if (*pos++ != ' ') {
+				pr_dbg4("invalid symbol file format for size\n");
+				continue;
+			}
+			type = *pos++;
+		}
+
 		if (*pos++ != ' ') {
 			pr_dbg4("invalid symbol file format after type\n");
 			continue;
@@ -959,7 +1012,8 @@ static int load_module_symbol_file(struct uftrace_symtab *symtab, const char *sy
 		if (type == ST_UNKNOWN || is_symbol_end(name)) {
 			if (symtab->nr_sym > 0) {
 				sym = &symtab->sym[symtab->nr_sym - 1];
-				sym->size = addr + offset - sym->addr;
+				if (sym->size == 0)
+					sym->size = addr + offset - sym->addr;
 			}
 			continue;
 		}
@@ -976,7 +1030,7 @@ static int load_module_symbol_file(struct uftrace_symtab *symtab, const char *sy
 		sym->addr = addr + offset;
 		sym->type = type;
 		sym->name = demangle(name);
-		sym->size = 0;
+		sym->size = size;
 
 		pr_dbg4("[%zd] %c %lx + %-5u %s\n", symtab->nr_sym, sym->type, sym->addr, sym->size,
 			sym->name);
@@ -1149,6 +1203,8 @@ void load_module_symtabs(struct uftrace_sym_info *sinfo)
 
 		map->mod = load_module_symtab(sinfo, map->libname, map->build_id);
 	}
+
+	load_python_symtab(sinfo);
 }
 
 /* returns the number of matching entries (1 = path only, 2 = build-id) */
@@ -1193,21 +1249,6 @@ int check_symbol_file(const char *symfile, char *pathname, int pathlen, char *bu
 	return ret;
 }
 
-static bool symbol_is_func(struct uftrace_symbol *sym)
-{
-	switch (sym->type) {
-	case ST_LOCAL_FUNC:
-	case ST_GLOBAL_FUNC:
-	case ST_WEAK_FUNC:
-	case ST_PLT_FUNC:
-	case ST_KERNEL_FUNC:
-		return true;
-
-	default:
-		return false;
-	}
-}
-
 char *make_new_symbol_filename(const char *symfile, const char *pathname, char *build_id)
 {
 	const char *p;
@@ -1234,8 +1275,7 @@ static void save_module_symbol_file(struct uftrace_symtab *stab, const char *pat
 {
 	FILE *fp;
 	unsigned i;
-	bool prev_was_plt = false;
-	struct uftrace_symbol *sym, *prev;
+	struct uftrace_symbol *sym;
 	char *newfile = NULL;
 
 	if (stab->nr_sym == 0)
@@ -1274,36 +1314,14 @@ static void save_module_symbol_file(struct uftrace_symtab *stab, const char *pat
 	if (strlen(build_id) > 0)
 		fprintf(fp, "# build-id: %s\n", build_id);
 
-	prev = &stab->sym[0];
-	prev_was_plt = (prev->type == ST_PLT_FUNC);
-
-	fprintf(fp, "%016" PRIx64 " %c %s\n", prev->addr - offset, (char)prev->type, prev->name);
-
 	/* PLT + normal symbols (in any order)*/
-	for (i = 1; i < stab->nr_sym; i++) {
+	for (i = 0; i < stab->nr_sym; i++) {
 		sym = &stab->sym[i];
 
-		/* mark end of the this kind of symbols */
-		if ((sym->type == ST_PLT_FUNC) != prev_was_plt) {
-			fprintf(fp, "%016" PRIx64 " %c __%ssym_end\n",
-				prev->addr + prev->size - offset, (char)ST_UNKNOWN,
-				prev_was_plt ? "dyn" : "");
-		}
-		else if (symbol_is_func(prev) && !symbol_is_func(sym)) {
-			fprintf(fp, "%016" PRIx64 " %c %s\n", prev->addr + prev->size - offset,
-				(char)ST_UNKNOWN, "__func_end");
-		}
-
-		fprintf(fp, "%016" PRIx64 " %c %s\n", sym->addr - offset, (char)sym->type,
-			sym->name);
-
-		prev = sym;
-		prev_was_plt = (prev->type == ST_PLT_FUNC);
+		fprintf(fp, "%016" PRIx64 " %08x %c %s\n", sym->addr - offset, sym->size,
+			(char)sym->type, sym->name);
 	}
 
-	fprintf(fp, "%016" PRIx64 " %c __%ssym_end\n", prev->addr + prev->size - offset,
-		(char)ST_UNKNOWN, prev_was_plt ? "dyn" : "");
-
 	fclose(fp);
 	free(newfile);
 }
