@@ -239,6 +239,7 @@ extern void setup_signal(void);
 		if (__ret < 0) {                                                                   \
 			pr_err("xasprintf");                                                       \
 		}                                                                                  \
+		__ret;                                                                             \
 	})
 
 #define xvasprintf(s, fmt, ap)                                                                     \
@@ -247,6 +248,7 @@ extern void setup_signal(void);
 		if (__ret < 0) {                                                                   \
 			pr_err("xvasprintf");                                                      \
 		}                                                                                  \
+		__ret;                                                                             \
 	})
 
 #define call_if_nonull(fptr, ...)                                                                  \
@@ -387,6 +389,8 @@ void strv_append(struct strv *strv, const char *str);
 void strv_replace(struct strv *strv, int idx, const char *str);
 char *strv_join(struct strv *strv, const char *delim);
 void strv_free(struct strv *strv);
+char *str_ltrim(char *str);
+char *str_rtrim(char *str);
 
 char **parse_cmdline(char *cmd, int *argc);
 void free_parsed_cmdline(char **argv);
